# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Requirement file container
~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import os as _os
import re as _re

from .. import _parse

logger = _logging.getLogger("deps.inspect.req_file")

REQUIREMENT = "requirement"
CONSTRAINT = "constraint"
EDITABLE = "editable"
INDEX = "index"


def tokens(filenames, base=None):
    """
    Walk over requirement tokens from a list of req files

    Parameters:
      filenames (iterable):
        List of filenames to parse

      base (str):
        Base path. If omitted or ``None``, the CWD is used.

    Returns:
      iterable: List of token objects found
    """
    base = _os.path.normpath(_os.path.abspath(base or _os.getcwd()))
    req_match = _re.compile(r"^(?:-r|--requirement)\s+(\S.*)").match
    con_match = _re.compile(r"^(?:-c|--constraint)\s+(\S+)").match
    edit_match = _re.compile(r"^(?:-e|--editable)\s+(\S.*)").match
    idx_match = _re.compile(r"^(?:-i|--index)\s+(\S+)").match
    name_match = _re.compile(r"^(?:%s)" % (_parse.regex.idre)).match

    stack = FileIterator(base, filenames)
    for line in stack:
        content = line.content.strip()
        if not content or content.startswith("#"):
            continue

        match = req_match(content)
        if match:
            req = match.group(1)
            logger.debug("Found requirement reference %r", req)
            if req.lower().startswith(("http://", "https://")):
                logger.debug("Skipping remote requirement file")
                continue

            if req.lower().startswith("file://"):
                req = req[len("file://") :]
            req = _os.path.normpath(
                _os.path.join(_os.path.dirname(line.file), req)
            )
            stack.push_file(req)
            continue

        match = con_match(content)
        if match:
            con = match.group(1)
            if con.lower().startswith("file://"):
                con = con[len("file://") :]
            if not con.startswith(("http://", "https://")):
                con = _os.path.normpath(
                    _os.path.join(_os.path.dirname(line.file), con)
                )

            logger.debug("Returning constraint file %r", con)
            yield Token.constraint(con.rstrip(), line)
            continue

        match = edit_match(content)
        if match:
            value = match.group(1).rstrip()
            logger.debug("Returning editable install %r", value)
            yield Token.editable(value, line)
            continue

        match = idx_match(content)
        if match:
            value = match.group(1).rstrip()
            logger.debug("Returning index %r", value)
            yield Token.index(value, line)
            continue

        if name_match(content):
            logger.debug("Returning requirement %r", content)
            yield Token.requirement(_parse.requirement(content), line)
            continue

        logger.debug(
            "Ignoring line %s in %s: %r", line.lineno, line.file, content
        )


class Token(object):
    """
    Container for parsed requirement file tokens

    Attributes:
      value (any):
        The token. Type depends on the token type. Typically a string, but
        it's an requirement object for type == REQUIREMENT

      line (Line):
        The line object

      type (str):
        The token type
    """

    # pylint: disable = redefined-builtin
    def __init__(self, value, line, type):
        """
        Initialization

        Parameters:
          value (any):
            The token value

          line (Line):
            The line

          type (str):
            The token type:
            - CONSTRAINT
            - REQUIREMENT
            - EDITABLE
            - INDEX
        """
        self.value = value
        self.line = line
        self.type = type

    def __repr__(self):
        """
        Create string representation

        Returns:
          str: String representation
        """
        return "%s(%r, line=%r, type=%s)" % (
            self.__class__.__name__,
            self.value,
            self.line,
            self.type.upper(),
        )

    @classmethod
    def constraint(cls, value, line):
        """
        Create CONSTRAINT token

        Parameters:
          value (str):
            The token value (constraint file reference)

          line (Line):
            The line

        Returns:
          Token: New Token instance
        """
        return cls(value, line, CONSTRAINT)

    @classmethod
    def requirement(cls, value, line):
        """
        Create REQUIREMENT token

        Parameters:
          value (parsed requirement):
            The token value (the parsed requirement object)

          line (Line):
            The line

        Returns:
          Token: New Token instance
        """
        return cls(value, line, REQUIREMENT)

    @classmethod
    def editable(cls, value, line):
        """
        Create EDITABLE token

        Parameters:
          value (str):
            The token value (editable reference, a folder or a URL)

          line (Line):
            The line

        Returns:
          Token: New Token instance
        """
        return cls(value, line, EDITABLE)

    @classmethod
    def index(cls, value, line):
        """
        Create INDEX token

        Parameters:
          value (str):
            The token value (the index URL)

          line (Line):
            The line

        Returns:
          Token: New Token instance
        """
        return cls(value, line, INDEX)


class FileIterator(object):
    """
    Iterator over file lines

    The file list can be extended while iterating over the lines. The file
    added latest will be processed first.
    """

    def __init__(self, base, filenames=None):
        """
        Initialization

        Parameters:
          base (str):
            The base path if relative file names are given

          filenames (iterable):
            The initial list of file names to iterate over
        """
        self._base = base
        self._stack = []
        self._seen = set()
        if filenames is not None:
            for filename in reversed(list(filenames)):
                self.push_file(filename)

    def push_file(self, filename):
        """
        Add a new file to iterate over

        The file will not be added if it has been seen already.

        Parameters:
          filename (str):
            The file name
        """
        filename = _os.path.normpath(_os.path.join(self._base, filename))
        key = _os.path.relpath(filename, self._base)
        if key in self._seen:
            logger.debug("Skipping file %r (already seen)", key)
            return
        self._seen.add(key)

        if not _os.path.isfile(filename):
            logger.debug("Skipping non-file %r", key)
            return

        logger.debug("Adding %r to the stack", key)
        lines = iter(File(filename))
        self._stack.append(getattr(lines, "__next__", None) or lines.next)

    def __iter__(self):
        """
        Iterate over lines

        Returns:
          iterable: All lines in LIFO (files) / FIFO (lines) order
        """
        while self._stack:
            try:
                yield self._stack[-1]()
            except StopIteration:
                self._stack.pop()


class File(object):
    """
    Requirement or constraints file

    Attributes:
      name (str):
        File name
    """

    def __init__(self, name):
        """
        Initialization

        Parameters:
          name (str):
            File name
        """
        self.name = name
        with open(self.name, "rb") as fp:
            self._lines = [
                Line(name, lineno, line.decode("latin-1"))
                for lineno, line in enumerate(fp, start=1)
            ]

    def __repr__(self):
        """
        Create string representation

        Returns:
          str: String representation
        """
        return "%s(%r)" % (
            self.__class__.__name__,
            self.name,
        )

    def __iter__(self):
        """
        Create line iterator

        Returns:
          iterable: Line iterator
        """
        return iter(self._lines)


class Line(object):
    """
    Line in a file

    Attributes:
      file (str):
        File name

      lineno (int):
        Line number

      content (str):
        The line
    """

    def __init__(self, file, lineno, content):
        """
        Initialization

        Parameters:
          file (str):
            File name

          lineno (int):
            Line number

          content (str):
            The line
        """
        self.file = file
        self.lineno = lineno
        self.content = content

    def __repr__(self):
        """
        Create string representation

        Returns:
          str: String representation
        """
        return "%s(%r, %r, %r)" % (
            self.__class__.__name__,
            self.file,
            self.lineno,
            self.content,
        )



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Upload packages
~~~~~~~~~~~~~~~

"""

import invoke as _invoke


@_invoke.task(default=True)
def source(ctx):
    """Upload source package"""
    files = list(ctx.shell.files("dist", "*.tar.gz"))
    if len(files) != 1:
        ctx.fail("Not exactly one tarball found")

    cmd = "twine upload --repository-url %s --username %s %s"
    args = [ctx.pypi.repository, ctx.pypi.username, files[0]]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd, *args), echo=True)


@_invoke.task()
def wheels(ctx):
    """Upload wheels"""
    files = list(ctx.shell.files("dist", "*.whl"))
    if not files:
        ctx.fail("No wheel found")

    cmd = (
        ctx.s("twine upload --repository-url")
        + [ctx.pypi.repository, "--username", ctx.pypi.username]
        + files
    )
    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Setup invoke namespace
~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import sys as _sys

from . import _default_settings
from . import shell as _shell
from . import tasks as _tasks
from . import util as _util

# pylint: disable = import-outside-toplevel


def setup():
    """
    Create invoke task namespace

    Returns:
      invoke.Collection: The main namespace
    """
    _logging.basicConfig(format="%(message)s")
    _logging.root.setLevel("ERROR")

    try:
        from .. import _settings
    except ImportError:
        settings = {}
    else:
        settings = _settings.settings

    env = {
        key: _util.adict(value) if isinstance(value, dict) else value
        for key, value in _util.dictmerge(
            _default_settings.default_settings(), settings
        ).items()
    }
    _sys.path.insert(0, _shell.root)

    return _tasks.setup_tasks(_util.adict(env))



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Git commands
~~~~~~~~~~~~

"""


def current_branch(ctx):
    """
    Find the current branch

    Returns:
      str: The current branch or None
    """
    return (
        command(
            ctx, ctx.s("branch --color=never --show-current"), hide=True
        ).stdout.strip()
        or None
    )


def sha(ctx):
    """
    Find current sha

    Returns:
      str: The current commit ID
    """
    return (
        command(ctx, ctx.s("log -1 --format=%H"), hide=True).stdout.strip()
        or None
    )


def command(ctx, cmd, **kwargs):
    """
    Run git command

    Parameters:
      cmd (list):
        The command to run

      **kwargs:
        Extra arguments for ctx.run()

    Returns:
      The run result
    """
    env = dict(LC_ALL="C")
    cmd = [ctx.which("git"), "--no-pager"] + cmd
    with ctx.shell.root_dir():
        return ctx.run(ctx.c(cmd), env=env, **kwargs)



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Build Tasks
~~~~~~~~~~~

"""

import os as _os

try:
    import build as _build
except ImportError:
    _build = None

import invoke as _invoke

from .._inv import tasks as _tasks
from . import _pkg_dir

# pylint: disable = import-outside-toplevel

NAMESPACE = "build"


@_invoke.task(default=True)
def source(ctx):
    """Build source package"""
    if source.avoid_rerun:
        return

    if _tasks.exists("doc.doc"):
        _tasks.execute(ctx, "doc.doc")

    with ctx.shell.root_dir():
        if not _os.path.isdir("setups"):
            _build_sdist(ctx)
            source.avoid_rerun = True
            return

    _source_multi(ctx)
    source.avoid_rerun = True


source.avoid_rerun = False


def _source_multi(ctx):
    """Build multiple source packages"""
    with ctx.shell.root_dir():
        ctx.shell.mkdir_p("dist")

        # Collect names
        names = sorted(
            set(
                name.rsplit(".", 1)[0]
                for name in _os.listdir("setups")
                if not name.startswith(".")
                and name.endswith((".py", ".toml"))
            )
        )

        # Build each package from a copy
        for name in names:
            with _pkg_dir.sdist(ctx, name) as pkg_root:
                _build_sdist(ctx)
                for archive in _os.listdir("dist"):
                    ctx.shell.cp(
                        _os.path.join(pkg_root, "dist", archive),
                        _os.path.join(ctx.shell.root, "dist", archive),
                    )


def _build_sdist(ctx):
    """Build single sdist in the CWD"""
    if ctx.get("wheels", {}).get("build") == "universal":
        with open("setup.cfg", "wb") as fp:
            fp.write(b"[bdist_wheel]\n")
            fp.write(b"universal = 1\n")

    if _build is None:
        ctx.run("python setup.py sdist")
    else:
        ctx.run("python -m build --sdist")



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared code for dependency files
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import re as _re

from ..._inv import shell as _shell
from .. import _parse
from . import _file_model

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.suggest.dep_file")


class DepFile(object):
    """
    Dependency file container

    Attributes:
      name (str):
        The filename
    """

    @property
    def type(self):
        """
        The dep file type

        :Type: str
        """
        return self._model.type

    def __init__(self, filename, model):
        """
        Initialization

        Parameters:
          filename (str):
            The filename
        """
        self.name = _shell.relative(filename)
        self._fullname = _shell.native(filename)
        self._model = model

    @classmethod
    def by_type(cls, type, filename):  # pylint: disable = redefined-builtin
        """
        Return dep file instance by type and name

        This factory returns the found initialized dep file ``None``.

        Parameters:
          type (str):
            The file type

          filename (str):
            The file name

        Returns:
          DepFile: new instance
        """
        # pylint: disable = used-before-assignment
        # pylint: disable = unnecessary-lambda-assignment
        find_subs = lambda cls: set(cls.__subclasses__()).union(
            s for c in cls.__subclasses__() for s in find_subs(c)
        )
        for sub in find_subs(_file_model.Model):
            if sub.type == type:
                return cls(filename, sub.by_name(filename))
        return None

    def __repr__(self):
        """
        Create debug representation

        Returns:
          str: The repr of the container
        """
        return "%s(%r, %r)" % (
            self.__class__.__name__,
            self.name,
            self._model,
        )

    def __hash__(self):
        """
        Hash function based on the filename

        Returns:
          int: The hash
        """
        return hash(self.name)

    def __eq__(self, other):
        """
        Equality - based on the name

        Returns:
          bool: Is it equal?
        """
        return isinstance(other, self.__class__) and other.name == self.name

    def __ne__(self, other):
        """
        Non-Equality - based on the name

        Returns:
          bool: Is it not equal?
        """
        return not self == other

    def __lt__(self, other):
        """
        Less-than comparision - based on the name

        Returns:
          bool: Is it smaller than other?
        """
        return isinstance(other, self.__class__) and self.name < other.name

    def __contains__(self, name):
        """
        Check if package is referenced

        Parameters:
          name (str):
            The package name to look for

        Returns:
          bool: Is it referenced?
        """
        for _ in self.find(name):
            return True
        return False

    def __str__(self):
        """
        The reconstructed file

        Returns:
          str: The file as string
        """
        return "".join(map(str, self._model.parts))

    def write(self):
        """Write back the file"""
        with open(self._fullname, "wb") as fp:
            fp.write(str(self).encode(self._model.encoding))

    def find(self, name, require_version=None):
        """
        Find requirements for `name`

        Parameters:
          name (str):
            The name to look for

          require_version (bool):
            Require the version to be present? If omitted or ``None``, the
            default is picked from the file model.

        Returns:
          iterable: Iterator over found specs
        """
        if require_version is None:
            require_version = self._model.require_version

        find = _re.compile(
            _parse.regex.full_spec(name, require_version=require_version)
        ).finditer

        for part in self._model.parts:
            if part.is_literal:
                continue
            for match in find(part.content):
                yield match.group("spec")

    def patch(self, patches):
        """
        Patch the file

        Parameters:
          patches (dict):
            The changeset to apply

        Returns:
          ReqFile: A new container instance with possibly patched content
        """
        parts = []
        for part in self._model.parts:
            if not part.is_literal:
                part = part.__class__(
                    part.raw, part.quote, self._patch(part.content, patches)
                )
            parts.append(part)
        return self.__class__(self._fullname, self._model.__class__(parts))

    @staticmethod
    def _patch(string, patches):
        """
        Patch a single string - helper

        Parameters:
          string (str):
            The string to patch

          patches (dict):
            Changeset (``{'changes': {name: {'new': 'spec'}, ...}}``)

        Returns:
          str: The patched string
        """

        def valid(req):
            """Check if it's valid to patch"""
            req = _parse.requirement(str(req))
            if req.marker and not req.marker.evaluate():
                return False
            return True

        result = string
        for name, patchlist in patches["changes"].items():
            for patch in patchlist:
                result = _re.sub(
                    _parse.regex.full_spec(name, require_version=True),
                    lambda m, n=patch["new"]: (
                        "%s%s" % (m.group("space"), n)
                        if valid(m.group("spec"))
                        else m.group(0)
                    ),
                    result,
                )

        return result



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Default settings
~~~~~~~~~~~~~~~~

"""

import sys as _sys

import invoke as _invoke

from . import shell as _shell
from . import util as _util

# pylint: disable = import-outside-toplevel


def fail(msg):
    """
    Exit with message

    Parameters:
      msg (str):
        The fail message

    Raises:
      invoke.Exit: Always raised
    """
    _sys.stderr.write("Error: %s\n" % (msg,))
    raise _invoke.Exit(code=1)


def which(executable):
    """
    Find executable - or fail

    Parameters:
      executable (str):
        The executable to look out for

    Returns:
      str: The full path

    Raises:
      invoke.Exit: If the executable was not found
    """
    found = _shell.frompath(executable)
    if not found:
        fail("%s not found" % (executable,))
    return found


def find_user():
    """
    Find system user

    Returns:
      str: The user's login name or ``None``
    """
    import getpass as _getpass

    try:
        return _getpass.getuser()
    except Exception:  # pylint: disable = broad-exception-caught
        return None


def default_settings():
    """
    Create default settings

    Returns:
      dict: Default settings
    """
    return dict(
        test=_util.adict(ignore=[]),
        shell=_util.adict(
            (key, value)
            for key, value in vars(_shell).items()
            if not key.startswith("_")
        ),
        c=_shell.command,
        q=lambda x: _shell.command("%s", x),
        s=_shell.split_command,
        fail=fail,
        which=which,
        user=find_user(),
        doc=_util.adict(
            userdoc="docs/userdoc",
            website=_util.adict(
                source="docs/website",
                target="dist/website",
            ),
            sphinx=_util.adict(
                want_apidoc=True,
                build="docs/_userdoc/_build",
                source="docs/_userdoc",
            ),
        ),
        pypi=_util.adict(
            # repository='https://test.pypi.org/legacy/',
            repository="https://upload.pypi.org/legacy/",
            username="__token__",
        ),
        paths=_util.adict(),
    )



# -*- coding: ascii -*-
"""
Project Settings
~~~~~~~~~~~~~~~~

"""


settings = dict(
    package="rjsmin",
    #
    # inv clean
    #
    clean=dict(
        cache=["bench/.tox/"],
    ),
    #
    # inv doc
    #
    doc=dict(
        sphinx=dict(want_apidoc=False),
    ),
    #
    # Wheels
    #
    wheels=dict(
        build="binary",
        specs={
            "aarch64": {
                "36": dict(manylinux="2014:2025.05.03-1", musllinux="1_1"),
                "37": dict(manylinux="2014:2025.05.03-1", musllinux="1_1"),
                "38": dict(manylinux="2014", musllinux="1_1"),
                "39": dict(manylinux="2014", musllinux="1_1"),
                "310": dict(manylinux="2014", musllinux="1_1"),
                "311": dict(manylinux="2014", musllinux="1_1"),
                "312": dict(manylinux="2014", musllinux="1_1"),
                "313": dict(manylinux="2014", musllinux="1_1"),
                "314": dict(manylinux="2014", musllinux="1_2"),
            },
            "x86_64": {
                "27": dict(manylinux="1"),
                "36": dict(manylinux="1", musllinux="1_1"),
                "37": dict(manylinux="1", musllinux="1_1"),
                "38": dict(manylinux="1", musllinux="1_1"),
                "39": dict(manylinux="1", musllinux="1_1"),
                "310": dict(manylinux="2010", musllinux="1_1"),
                "311": dict(manylinux="2014", musllinux="1_1"),
                "312": dict(manylinux="2014", musllinux="1_1"),
                "313": dict(manylinux="2014", musllinux="1_1"),
                "314": dict(manylinux="2014", musllinux="1_2"),
            },
            "i686": {
                "27": dict(manylinux="1"),
                "36": dict(manylinux="1", musllinux="1_1"),
                "37": dict(manylinux="1", musllinux="1_1"),
                "38": dict(manylinux="1", musllinux="1_1"),
                "39": dict(manylinux="1", musllinux="1_1"),
                "310": dict(manylinux="2010", musllinux="1_1"),
                "311": dict(manylinux="2014", musllinux="1_1"),
                "312": dict(manylinux="2014", musllinux="1_1"),
                "313": dict(manylinux="2014", musllinux="1_1"),
                "314": dict(manylinux="2014", musllinux="1_2"),
            },
        },
    ),
)



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test suite tasks
~~~~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from . import _features
from . import pypi as _pypi
from ._inv import tasks as _tasks


@_tasks.optional(None, _features.python_tests)
@_invoke.task()
def local(
    ctx,
    new_first=False,
    failed_first=False,
    keep_going=False,
    capture=False,
    coverage=False,
    coverage_report=True,
    only=None,
):
    """
    Run the unit test suite using py.test

    Parameters:
      new_first (bool):
        Run tests from new files first. Default: false

      failed_first (bool):
        Run all tests, but run the last failures first. Default: false

      keep_going (bool):
        Keep going after a failure? Default: false

      capture (bool):
        Capture output? Default: false

      coverage (bool):
        Measure test coverage? Default: false

      coverage_report (bool):
        Emit coverage report? Only relevant if coverage is enabled.
        Default: true

      only (str):
        Filter expression for tests to run
    """
    # pylint: disable = too-many-positional-arguments

    cmd = [ctx.which("py.test")] + ctx.s(
        "-vv --color=yes --doctest-modules --doctest-ignore-import-errors"
    )
    if new_first:
        cmd += ["--nf"]
    if failed_first:
        cmd += ["--ff"]
    if not keep_going:
        cmd += ["--exitfirst"]
    if not capture:
        cmd += ["-s"]
    if coverage:
        cmd += ["--no-cov-on-fail", "--cov", ctx.package]
        if coverage_report:
            cmd += ["--cov-report=html", "--cov-report=term"]
        else:
            cmd += ["--cov-report="]
    if only:
        cmd += ["-k", only]

    for ignored in ctx.test.ignore:
        cmd += ["--ignore", ignored]

    # cmd += [ctx.package.replace(".", "/")]
    cmd += ["tests"]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(None, _features.tox_tests)
@_invoke.task(default=True)
def tox(ctx, rebuild=False, env=None, hashseed=None):
    """
    Run the test suite using tox

    Parameters:
      rebuild (bool):
        Rebuild tox environment(s)? Default: false

      env (str):
        A single test environment to run. All environments are run otherwise.

      hashseed (str):
        The hashseed to use. Only specify as a debugging measure.
    """
    cmd = [ctx.which("tox")]
    if rebuild:
        cmd += ["-r"]
    if env is not None:
        cmd += ["-e", env]
    if hashseed is not None:
        cmd += ["--hashseed", hashseed]

    with ctx.shell.root_dir():
        ctx.run(
            ctx.c(cmd),
            env=dict(_os.environ, PIP_INDEX_URL=_pypi.index_url(ctx)),
            echo=True,
        )



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Cleanup tasks
~~~~~~~~~~~~~

"""

import invoke as _invoke

from ._inv import tasks as _tasks


@_invoke.task()
def py(ctx):
    """Wipe *.py[co] files"""
    ctx.shell.rm_rf(ctx.shell.glob("**/__pycache__/", "**/*.py[co]"))


@_invoke.task(py)
def dist(ctx):
    """Wipe all"""
    _tasks.execute(ctx, "clean.clean", so=True, cache=True)
    ctx.shell.rm_rf(ctx.shell.glob("*.egg-info"))


@_invoke.task(py, default=True)
def clean(ctx, so=False, cache=False):
    """
    Wipe *.py[co] files and test leftovers

    Parameters:
      so (bool):
        Clean *.so files, too? Default: false

      cache (bool):
        Clean various cache files and directories, too? Default: false
    """
    ctx.shell.rm(ctx.shell.files(".", ".coverage*"))
    path = ctx.paths.get("terraform")
    if path:
        gpath = ctx.shell.glob_escape(path.rstrip("/")) + "/**/"
        ctx.shell.rm_rf(ctx.shell.glob(gpath + "*.auto.tfvars.json"))
        ctx.shell.rm_rf(ctx.shell.glob(gpath + "*.tf.json"))
    path = ctx.paths.get("cloudformation")
    if path:
        gpath = ctx.shell.glob_escape(path.rstrip("/")) + "/**/"
        ctx.shell.rm_rf(ctx.shell.glob(gpath + "*.auto.*.yaml"))
        ctx.shell.rm_rf(ctx.shell.glob(gpath + "*.zip"))

    ctx.shell.rm_rf(ctx.shell.glob("tasks/**/*.generated*"))
    ctx.shell.rm_rf(
        "gcov.out",
        "setup.cfg",
        "docs/coverage/",
        "docs/gcov/",
        "build/",
        "dist/",
        ".sdist/",
        ".bdist/",
        ctx.doc.userdoc,
        ctx.doc.sphinx.build,
        ctx.doc.website.source,
        ctx.doc.website.target,
    )
    more = ctx.get("clean", {}).get("default", [])
    for glob in more:
        ctx.shell.rm_rf(ctx.shell.glob(glob))

    if cache:
        _tasks.execute(ctx, "clean.cache")
    if so:
        _tasks.execute(ctx, "clean.so")


@_invoke.task(name="cache")
def cacheclean(ctx):
    """Wipe Cache files"""
    path = ctx.paths.get("terraform")
    if path:
        gpath = ctx.shell.glob_escape(path.rstrip("/")) + "/**/"
        ctx.shell.rm_rf(ctx.shell.glob(gpath + ".terraform/"))
        ctx.shell.rm_rf(ctx.shell.glob(gpath + ".terraform.lock.hcl"))

    ctx.shell.rm_rf(
        ".tox/",
        ".cache/",
        ".pytest_cache/",
        "tests/.cache/",
        "tests/.pytest_cache/",
        ".mypy_cache/",
    )
    more = ctx.get("clean", {}).get("cache", [])
    for glob in more:
        ctx.shell.rm_rf(ctx.shell.glob(glob))


@_invoke.task(name="so")
def soclean(ctx):
    """Wipe *.so files"""
    ctx.shell.rm(ctx.shell.glob("**/*.pyd", "**/*.so"))



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Propose updates to current dependencies
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import collections as _collections
import logging as _logging

from . import _parse
from . import _suggest

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.patch")


def patch_updates(ctx, config, upgrade=False, verbose=True, debug=False):
    """
    Patch requirement files for dependency updates

    Parameters:
      config (adict):
        The deps config

      upgrade (bool):
        Consider upgrades?

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false

    Returns:
      dict: Changelog (``{bucket -> {name -> new version}}``)
    """
    logger.setLevel("DEBUG" if debug else ("INFO" if verbose else "WARN"))

    changes = _suggest.suggest_updates(
        ctx, config, upgrade=upgrade, verbose=verbose, debug=debug
    )
    replace, latest = changes["replace"], changes["latest"]
    if latest and not latest.as_patch_info():
        latest = None
    if not replace and not latest:
        logger.info("Nothing to do.")
        return {}

    changelog = _collections.defaultdict(
        lambda: _collections.defaultdict(dict)
    )
    boilerplate = set(_parse.normalize(item) for item in config.boilerplate)

    for file, change in sorted(replace.items()):
        logger.info("Patching %s...", file.name)
        file.patch(change).write()

        for name, update in change["changes"].items():
            bucket = (
                "boilerplate"
                if name in boilerplate or file.type == "text"
                else "update"
            )
            changelog[bucket][name] = sorted(
                set(item["new_version"] for item in update)
            )

    if latest:
        # The file might have been modified, so reload
        error = latest.load()
        if error:
            ctx.fail(error)

        latest.write()
        changelog["support"] = latest.as_changed()
        for name in changelog["support"]:
            changelog["update"].pop(name, None)

    return {key: dict(value) for key, value in changelog.items() if value}



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Propose updates to current dependencies
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import collections as _collections
import itertools as _it
import logging as _logging
import os as _os

from .. import _inspect
from .. import _parse
from . import _dep_file
from . import _latest_file
from . import _requirement

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.suggest")


def suggest_updates(ctx, config, upgrade=False, verbose=True, debug=False):
    """
    Inspect dependencies

    Parameters:
      config (adict):
        The deps config

      upgrade (bool):
        Consider upgrades?

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false

    Returns:
      dict: {"latest": dict, "replace": dict}
    """
    # pylint: disable = too-many-locals

    logger.setLevel("DEBUG" if debug else ("INFO" if verbose else "WARN"))

    changes = _collections.defaultdict(list)
    candidates = {
        key: val
        for key, val in _inspect.inspect_dependencies(
            ctx,
            config,
            verbose=verbose,
            debug=debug,
        ).items()
        if val["updates"]
    }
    if not candidates:
        logger.debug("No dependencies found to update")
        return {"latest": {}, "replace": {}}

    latest = {}
    not_latest = set(_parse.normalize(item) for item in config.no_latest)
    for info in _with_locations(ctx, candidates).values():
        for loc, want_upgrade, compat in _updatable(
            config, upgrade, info["locations"]
        ):
            oreq = str(loc["requirement"])
            req, version = _replace(loc, info, want_upgrade, compat)
            preq = _parse.requirement(req)

            if compat and _parse.normalize(preq.name) not in not_latest:
                lreq, lversion = _replace(
                    loc, info, want_upgrade, False, latest=True
                )
                plreq = _parse.requirement(lreq)
                key = _parse.normalize(plreq.name)

                if str(plreq) == str(preq):
                    not_latest.add(key)
                elif key not in not_latest:
                    latest[key] = dict(
                        new=lreq,
                        new_version=lversion,
                    )

            if str(preq) == oreq:
                logger.debug("Unchanged requirement: %r", req)
                continue

            changes[loc["file"]].append(
                dict(
                    new_version=version,
                    new=req,
                    old=loc["content"],
                )
            )

    # pylint: disable = unnecessary-lambda-assignment
    as_key = lambda x: _parse.normalize(_parse.requirement(x).name)
    group_key = lambda x: as_key(x["old"])

    return dict(
        latest=_suggest_latest(config, latest, not_latest),
        replace={
            file: dict(
                changes={
                    name: list(group)
                    for name, group in _it.groupby(
                        sorted(changelist, key=group_key), key=group_key
                    )
                }
            )
            for file, changelist in changes.items()
        },
    )


def _suggest_latest(config, latest, not_latest):
    """
    Suggest latest deps for packages

    Parameters:
      config (adict):
        The deps config

      latest (dict):
        Package -> version info

     not_latest (set):
        Set of packages not to be added

    Returns:
      dict: {add: [str], remove: [str]}
    """
    return (
        _latest_file.LatestFile(
            config.latest.file,
            config.latest.pattern,
            config.latest.prefix,
            config.latest.suffix,
            latest,
            not_latest,
        )
        if config.compat
        else None
    )


def _replace(loc, info, upgrade, compat, latest=False):
    """
    Find replacement for a single requirement

    Parameters:
      loc (dict):
        Location

      info (dict):
        Package update info

      upgrade (bool):
        Try upgrading to the latest version?

      compat (bool):
        Ensure compatible changes? (``~=``)

      latest (bool):
        The target is the list of "latest" requirements (e.g. in
        development.txt)

    Returns:
      tuple: new requirement string and new version (extracted)
             (``(str, str)``)
    """
    req, version = _requirement.replace(
        loc["content"],
        info["updates"],
        upgrade=upgrade,
        compat=compat,
        latest=latest,
    )
    logger.debug(
        "upgrade: %r, compat: %r, latest: %r, req: %s, v: %r",
        upgrade,
        compat,
        latest,
        req,
        version,
    )
    return req, version


def _updatable(config, want_upgrade, locations):
    """
    Find and classify updatable locations

    Parameters:
      config (adict):
        The deps config

      want_upgrade (bool):
        Upgrade requested?

      locations (iterable):
        Locations to consider

    Returns:
      iterable: of (location, upgrade possible, compat required)
    """
    no_compat = set(_parse.normalize(item) for item in config.no_compat)
    no_upgrade = set(_parse.normalize(item) for item in config.no_upgrade)

    for loc in locations:
        if (
            loc["requirement"].marker
            and not loc["requirement"].marker.evaluate()
        ):
            logger.debug("Skipping %r, because it doesn't evaluate", loc)
            continue

        if loc["file"].name in config.keep_as_is:
            logger.debug("Skipping %r (keeping as-is)", loc)
            continue

        want_no_upgrade = (
            _parse.normalize(loc["requirement"].name) in no_upgrade
        )
        upgrade = False if want_no_upgrade else want_upgrade

        if not want_no_upgrade and loc["file"].name in config.upgrade:
            compat = False
            upgrade = True
        elif _parse.normalize(loc["requirement"].name) in no_compat:
            compat = False
        else:
            compat = loc["file"].type != "text" and config.compat

        yield loc, upgrade, compat


def _with_locations(ctx, candidates):
    """
    Setup candidate locations

    Parameters:
      candidates (dict):
        The candidates to prepare

    Returns:
      dict: candidates again
    """
    extra = []

    with ctx.shell.root_dir():
        if _os.path.isfile("setup.py"):
            extra.append(_dep_file.DepFile.by_type("setup", "setup.py"))
            logger.debug("Extra location found: %r", extra[-1])

        if _os.path.isfile("pyproject.toml"):
            extra.append(_dep_file.DepFile.by_type("toml", "pyproject.toml"))
            logger.debug("Extra location found: %r", extra[-1])

        if _os.path.isdir("setups"):
            for name in sorted(_os.listdir("setups")):
                if name.startswith("."):
                    continue

                if name.endswith(".py"):
                    extra.append(
                        _dep_file.DepFile.by_type(
                            "setup", "setups/%s" % (name,)
                        )
                    )
                    logger.debug("Extra location found: %r", extra[-1])

                elif name.endswith(".toml"):
                    extra.append(
                        _dep_file.DepFile.by_type(
                            "toml", "setups/%s" % (name,)
                        )
                    )
                    logger.debug("Extra location found: %r", extra[-1])

    for name, info in candidates.items():
        info["locations"] = [
            dict(
                file=_dep_file.DepFile.by_type("text", item["file"]["name"]),
                content=item["file"]["content"].strip(),
                requirement=item["requirement"],
            )
            for item in info["locations"]
            if item.get("file")
        ] + [
            dict(
                file=depfile,
                content=req,
                requirement=_parse.requirement(req),
            )
            for depfile in extra
            for req in depfile.find(name)
        ]

    return candidates



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dependency Management Tasks
~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import json as _json
import os as _os
import sys as _sys

import invoke as _invoke

from .. import pypi as _pypi
from .._inv import tasks as _tasks
from .._inv import util as _util

# pylint: disable = import-outside-toplevel

NAMESPACE = "deps"


def _default_config(ctx):
    """
    Set default config

    Returns:
      adict: ctx.deps after defaults being applied
    """
    if getattr(_default_config, "applied", None):
        return ctx.deps

    ctx.deps = _util.dictmerge(
        _util.adict(
            toplevel=[
                "development.txt",
                "tests/requirements.txt",
            ],
            upgrade=[
                "checkout-requirements.txt",
            ],
            keep_as_is=[
                "compat-requirements.txt",
            ],
            no_compat=["setuptools", "pip", "build"],
            no_upgrade=[],
            no_latest=[],
            boilerplate=["setuptools", "pip", "build"],
            compat=True,
            latest=_util.adict(
                file="development.txt",
                pattern=r"(?=^-e\s+\.)",
                prefix="# Latest dependencies",
                suffix="",
            ),
        ),
        ctx.get("deps", {}),
    )
    _default_config.applied = True
    return ctx.deps


@_invoke.task()
def old(ctx):
    """List outdated python packages"""
    with ctx.shell.root_dir():
        ctx.run(ctx.c("pip list -i %s -o", _pypi.index_url(ctx)), echo=True)


@_invoke.task()
def package(ctx, upgrade=False):
    """
    Update python dependencies, excluding development (``-e .``)

    Parameters:
      upgrade (bool):
        Run pip install with ``-U`` flag?
    """
    cmd = ["pip", "install", "-i", _pypi.index_url(ctx)]
    if upgrade:
        cmd += ["-U"]
    cmd += ["-e", "."]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_invoke.task(default=True)
def dev(ctx, upgrade=False):
    """
    Update python dependencies, including development (``-r development.txt``)

    Parameters:
      upgrade (bool):
        Run pip install with ``-U`` flag?
    """
    cmd = ["pip", "install", "-i", _pypi.index_url(ctx)]
    if upgrade:
        cmd += ["-U"]
    cmd += ["-r", "development.txt"]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_invoke.task()
def reset(ctx, upgrade=False):
    """
    Reset your virtual env

    This command uninstalls everything except editable installs and reinstalls
    from scratch (``-r development.txt``)

    Parameters:
      upgrade (bool):
        Run pip install with ``-U`` flag?
    """
    cmd = [ctx.which("bash"), "-il", "%s/reset.sh"]
    if upgrade:
        cmd += ["-u"]
    cmd += ["."]
    with ctx.shell.root_dir():
        ctx.run(
            ctx.c(cmd, ctx.shell.native(_os.path.dirname(__file__))),
            pty=True,
        )


@_invoke.task()
def inspect(ctx, deps=False, verbose=False, debug=False):
    """
    Inspect current dependencies and print to stdout

    Parameters:
      deps (bool):
        run ``inv deps`` before?

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false
    """
    if deps:
        _tasks.execute(ctx, "deps.dev")

    from . import _inspect

    _sys.stdout.write(
        _json.dumps(
            _inspect.inspect_dependencies(
                ctx,
                _default_config(ctx),
                verbose=verbose,
                debug=debug,
            ),
            indent=4,
            default=str,
        )
        + "\n"
    )


@_invoke.task()
def check(ctx, upgrade=False, deps=False, verbose=True, debug=False):
    """
    Suggest dependency changes (basically dry run for `inv deps.patch`)

    Parameters:
      deps (bool):
        run ``inv deps`` before?

      upgrade (bool):
        Consider upgrades? They might be incompatible with your code.
        Default: false

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false
    """
    if deps:
        _tasks.execute(ctx, "deps.dev")

    from . import _suggest

    result = _suggest.suggest_updates(
        ctx,
        _default_config(ctx),
        upgrade=upgrade,
        verbose=verbose,
        debug=debug,
    )

    result["replace"] = {
        file.name: dict(type=file.type, **info)
        for file, info in result["replace"].items()
    }
    result["latest"] = result["latest"] and result["latest"].as_patch_info()
    _sys.stdout.write(
        _json.dumps(
            result,
            indent=4,
            default=str,
        )
        + "\n"
    )


@_invoke.task()
def patch(ctx, deps=False, upgrade=False, verbose=True, debug=False):
    """
    Find dependency updates and patch the files

    Parameters:
      deps (bool):
        run ``inv.deps`` before?

      upgrade (bool):
        Consider upgrades? They might be incompatible with your code.
        Default: false

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false
    """
    if deps:
        _tasks.execute(ctx, "deps.dev")

    from . import _patch

    _sys.stdout.write(
        _json.dumps(
            _patch.patch_updates(
                ctx,
                _default_config(ctx),
                upgrade=upgrade,
                verbose=verbose,
                debug=debug,
            ),
            indent=4,
            default=str,
        )
        + "\n"
    )



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Regular expressions
~~~~~~~~~~~~~~~~~~~

:See: https://www.python.org/dev/peps/pep-0508/
"""
import functools as _ft
import re as _re

# pylint: disable = import-outside-toplevel


class regex(object):
    """Collection of regexes"""

    wsp = r"[^\S\r\n]"
    idre = r"([a-zA-Z\d]+(?:[._-]+[a-zA-Z\d]+)*)"
    extra = r"""(?: \[
        %(wsp)s* %(idre)s+ %(wsp)s*
        (?:,%(wsp)s* %(idre)s %(wsp)s*)*
    \] )""" % dict(
        idre=idre, wsp=wsp
    )

    version_cmp = r"(?:===|[<!=>~]=|[<>])"
    version_str = r"[a-zA-Z0-9_.*+!-]+"
    version_release = r"""(?x)(?:
        ([vV]? (?:[0-9]+!)?)
        ([0-9]+(?:\.[0-9]+)*)
    )"""
    version1 = r"(?: %(cmp)s %(wsp)s* %(version)s )" % dict(
        cmp=version_cmp, wsp=wsp, version=version_str
    )
    version = r"""(?:
        %(version1)s %(wsp)s*
        (?:, %(wsp)s* %(version1)s %(wsp)s* )*
    )""" % dict(
        version1=version1, wsp=wsp
    )

    marker = r"""(?:
        (?: ; %(wsp)s*
            (?:
                [a-zA-Z0-9"'<>=~!().{}_*#\\:,/?[\]`@$%%^&+|-]+
                %(wsp)s*
            )+
        )+
    )""" % dict(
        wsp=wsp
    )

    # the # character is enclosed in a character class, so it can be used
    # with re.X
    line_comment = r"(?:[#][^\r\n]*)"

    py_str_double = r'(?:"[^"\\\r\n]*(?:\\(?:[^\r\n]|\r?\n|\r)[^"\\\r\n]*)*")'
    py_mstr_double = r'(?:"""[^"\\]*(?:(?:\\.|""?[^"])[^"\\]*)*""")'
    py_str_single = r"(?:'[^'\\\r\n]*(?:\\(?:[^\r\n]|\r?\n|\r)[^'\\\r\n]*)*')"
    py_mstr_single = r"(?:'''[^'\\]*(?:(?:\\.|''?[^'])[^'\\]*)*''')"

    toml_str_double = py_str_double
    toml_mstr_double = (
        r'(?:"""[^"\\]*(?:(?:\\.|""(?!"""")|"(?!"""""))[^"\\]*)*""")'
    )
    toml_str_single = r"(?:'[^'\r\n]*(?:[^'\r\n]*)*')"
    toml_mstr_single = r"(?:'''[^']*(?:''?[^'][^']*)*''')"

    @classmethod
    def full_spec(cls, name=None, require_version=False):
        """
        Full version spec regex

        Parameters:
          name (str):
            The package name to match against. If omitted or ``None`` all
            valid names are matched

          require_version (bool):
            Require the version to be present?

        Returns:
          str: The regex
        """
        return r"""(?xm)
            (?P<space> %(wsp)s* )
            (?P<spec>
                (?P<name>    %(name)s                                   )
                (?P<extra>   (?: %(wsp)s* %(extra)s )   ?               )
                (?P<version> (?: %(wsp)s* %(version)s ) %(opt_version)s )
                (?P<marker>  (?: %(wsp)s* %(marker)s )  ?               )
            )
        """ % dict(
            wsp=cls.wsp,
            name=cls.idre if name is None else name_as_regex(name),
            extra=cls.extra,
            version=cls.version,
            opt_version="" if require_version else "?",
            marker=cls.marker,
        )


def name_as_regex(name):
    """
    Transform package name into a regex

    Letters will be matched case insensitively and interpunction chars will be
    matched as a sequence.

    >>> name_as_regex("foo-b4r")
    '(?<![a-zA-Z0-9_.-])[fF][oO][oO][_.-]+[bB]4[rR](?![a-zA-Z0-9_.-])'

    Parameters:
     name (str):
        The name to transform

    Returns:
      str: The name as regex
    """
    # pylint: disable = unnecessary-lambda-assignment

    # escape control characters
    name = _re.sub(r"([^a-zA-Z0-9_.-])", r"\\\1", name)

    # replace interpunction sequences with matches of interpunction sequences
    name = _re.sub(r"[_.-]+", "[_.-]+", name)

    # match all letters CI
    ci_match = lambda m: "[%s%s]" % (m.group(1).lower(), m.group(1).upper())
    name = _re.sub(r"([a-zA-Z])", ci_match, name)

    # make sure we're matching the full name
    name = r"(?<![a-zA-Z0-9_.-])%s(?![a-zA-Z0-9_.-])" % (name,)

    return name


# group(1) is the match, group(2) the comment (if it is a comment match)
find_py_str_or_comment = _re.compile(
    r"""(?s) (
          (%(line_comment)s)
        | %(py_mstr_double)s
        | %(py_mstr_single)s
        | %(py_str_double)s
        | %(py_str_single)s
    ) """
    % vars(regex),
    _re.X,
).finditer

find_toml_str_or_comment = _re.compile(
    r"""(?s) (
          (%(line_comment)s)
        | %(toml_mstr_double)s
        | %(toml_mstr_single)s
        | %(toml_str_double)s
        | %(toml_str_single)s
    ) """
    % vars(regex),
    _re.X,
).finditer


safe_extra_sub = _ft.partial(_re.compile(r"[^A-Za-z0-9.-]+").sub, "_")
name_clean_sub = _ft.partial(_re.compile(r"[-_.]+").sub, "-")


def safe_extra(value):
    """
    Convert string to safe extra name

    Parameters:
      value (str):
        The string to sanitize

    Returns:
      str: The safe extra
    """
    return safe_extra_sub(value).lower()


def normalize(name):
    """
    Normalize package name

    Parameters:
      name (str):
        The name to normalize

    Returns:
      str: The normalized name
    """
    return name_clean_sub(name).lower()


def _make_req_parse():
    """
    Pick requirement parser

    Returns:
      callable: Parser call
    """
    try:
        import packaging.requirements
    except ImportError:
        import pkg_resources
    else:
        return packaging.requirements.Requirement
    return pkg_resources.Requirement.parse


requirement = _make_req_parse()


def _make_ver_parse():
    """
    Pick version parser

    Returns:
      callable: Parser call
    """
    try:
        import packaging.version
    except ImportError:
        import pkg_resources
    else:
        return packaging.version.Version
    return pkg_resources.parse_version


version = _make_ver_parse()



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Release code
~~~~~~~~~~~~

"""

import re as _re


def check_committed(ctx):
    """Check if everything is committed"""
    git = ctx.shell.frompath("git")
    lines = ctx.run(
        ctx.c("%s branch --color=never", git), env=dict(LC_ALL="C"), hide=True
    ).stdout.splitlines()
    for line in lines:
        if line.startswith("*"):
            branch = line.split(None, 1)[1]
            break
    else:
        ctx.fail("Could not determine current branch.")

    if branch != "master":
        rex = _re.compile(r"^\d+(?:\.\d+)*\.[xX]$").match
        match = rex(branch)
        if not match:
            ctx.fail("Not in master or release branch.")

    lines = ctx.run(
        ctx.c("%s status --porcelain", git), env=dict(LC_ALL="C"), hide=True
    ).stdout
    if lines:
        ctx.fail("Uncommitted changes!")


def add_tag(ctx):
    """Add release tag"""
    version = ctx.run("python setup.py --version", hide=True).stdout.strip()
    git = ctx.shell.frompath("git")

    ctx.run(
        ctx.c(
            """
        %s tag -a -m "Release version %s" -- %s
    """,
            git,
            version,
            version,
        ),
        echo=True,
    )



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Release code
~~~~~~~~~~~~

"""

import errno as _errno
import os as _os


def update(ctx):
    """Update version in relevant places"""
    version = ctx.run("python setup.py --version", hide=True).stdout.strip()

    _userdoc(ctx, version)
    _download(ctx, version)
    _changes(ctx, version)


def _userdoc(ctx, version):
    """Update version in userdoc"""
    short = ".".join(version.split(".")[:2])
    conf = _os.path.join(ctx.doc.sphinx.source, "conf.py")

    with open(conf, "rb") as fp:
        lines = fp.read().decode("latin-1").splitlines(True)

    seen = set()
    with open(conf, "wb") as fp:
        for line in lines:
            if line.startswith("version"):
                line = "version = %r\n" % (short,)
                seen.add("version")
            elif line.startswith("release"):
                line = "release = %r\n" % (version,)
                seen.add("release")
            fp.write(line.encode("latin-1"))

    assert len(seen) == 2, "version/release not found in userdoc/conf.py"


def _download(ctx, version):  # noqa
    """Update version in download files"""
    # pylint: disable = too-many-branches, too-many-statements

    filename = _os.path.join(ctx.doc.sphinx.source, "website_download.txt")
    isdev = "dev" in version

    dllines, dlpath = [], ""
    if isdev:
        oldstable, hasstable = [], False
        try:
            with open(filename, "rb") as fp:
                lines = fp.read().decode("latin-1").splitlines(True)
        except IOError as e:
            if e.errno != _errno.ENOENT:
                raise
        else:
            for line in lines:
                if line.startswith(".. begin stable"):
                    hasstable, dllines = True, oldstable
                oldstable.append(line)
        if not hasstable:
            dlpath = "dev/"

    newdev = []
    with open(filename + ".in", "rb") as fp:
        lines = fp.read().decode("latin-1").splitlines(True)
    if not dllines:
        dllines = lines
    else:
        for line in lines:
            if newdev:
                newdev.append(line)
                if line.startswith(".. end dev"):
                    break
            elif line.startswith(".. begin dev"):
                newdev.append(line)
        else:
            ctx.fail(
                "Incomplete/missing dev marker in %s" % (filename + ".in",)
            )

    instable, indev = [], []
    with open(filename, "wb") as fp:
        for line in dllines:
            if instable:
                instable.append(line)
                if line.startswith(".. end stable"):
                    if isdev:
                        res = "".join(instable) if hasstable else ""
                    else:
                        res = (
                            "".join(instable)
                            .replace("@@VERSION@@", version)
                            .replace("@@PATH@@", "")
                        )
                    fp.write(res.encode("latin-1"))
                    instable = []
            elif indev:
                indev.append(line)
                if line.startswith(".. end dev"):
                    if not isdev:
                        res = "".join([indev[0], indev[-1]])
                    else:
                        res = (
                            "".join(newdev or indev)
                            .replace("@@DEVVERSION@@", version)
                            .replace("@@PATH@@", "dev/")
                        )
                    fp.write(res.encode("latin-1"))
                    indev = []
            elif line.startswith(".. begin stable"):
                instable.append(line)
            elif line.startswith(".. begin dev"):
                indev.append(line)
            elif isdev and hasstable:
                fp.write(line.encode("latin-1"))
            else:
                fp.write(
                    line.replace("@@VERSION@@", version)
                    .replace("@@PATH@@", dlpath)
                    .encode("latin-1")
                )


def _changes(ctx, version):
    """Update version in CHANGES"""
    fname = ctx.shell.native("CHANGES")
    with open(fname, "rb") as fp:
        lines = fp.read().decode("latin-1").splitlines(True)

    with open(fname, "wb") as fp:
        for line in lines:
            if line.rstrip() == "Changes with version":
                line = "%s %s\n" % (line.rstrip(), version)
            fp.write(line.encode("latin-1"))



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Run code formatters
~~~~~~~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from . import _features
from ._inv import tasks as _tasks

_FORMATTERS = []


@_tasks.optional(_FORMATTERS, _features.isort)
@_invoke.task()
def isort(ctx, diff=False):
    """
    Format python code using isort formatter

    Parameters:
      diff (bool):
        Just emit a diff instead of changing files inline? Default: false
    """
    cmd = [ctx.which("isort")]
    if diff:
        cmd += ctx.s("--diff --color")
    cmd += ctx.s("--settings-path pyproject.toml")
    cmd += ["tasks"]
    if ctx.get("package"):
        cmd += [ctx.package]
    if _os.path.exists(ctx.shell.native("tests")):
        cmd += ["tests"]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_FORMATTERS, _features.black)
@_invoke.task()
def black(ctx, diff=False):
    """
    Format python code using Black formatter

    Parameters:
      diff (bool):
        Just emit a diff instead of changing files inline? Default: false
    """
    cmd = [ctx.which("black")]
    if diff:
        cmd += ctx.s("--diff --color")
    cmd += ctx.s("--config pyproject.toml .")

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_FORMATTERS, _features.terraform)
@_invoke.task(name="tf")
def terraform(ctx):
    """Format terraform code"""
    base_cmd = [ctx.which("terraform"), "-chdir=" + ctx.paths.terraform]
    with ctx.shell.root_dir():
        ctx.run(ctx.c(base_cmd + ctx.s("fmt -recursive")), echo=True)


@_tasks.optional(_FORMATTERS)
@_invoke.task(name="all", default=True)
def all_(ctx):
    """Run all formatters"""
    _tasks.execute(ctx, *["format." + formatter for formatter in _FORMATTERS])



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Run code checkers
~~~~~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from . import _features
from ._inv import tasks as _tasks

_CHECKERS = []
_CHECKERS_TF = []
_CHECKERS_SAM = []


@_tasks.optional(_CHECKERS, _features.pylint)
@_invoke.task("clean.py")
def lint(ctx):
    """Run pylint"""
    if ctx.get("package"):
        cmd = (
            [ctx.which("pylint")] + ctx.s("--rcfile pylintrc") + [ctx.package]
        )
        with ctx.shell.root_dir():
            ctx.run(ctx.c(cmd), echo=True)

    cmd = [ctx.which("pylint")] + ctx.s("--rcfile tasks/pylintrc") + ["tasks"]
    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_CHECKERS, _features.mypy)
@_invoke.task("clean.py")
def mypy(ctx):
    """Run mypy"""
    cmd = [
        ctx.which("mypy"),
        "--config-file",
        "pyproject.toml",
        "-p",
        ctx.package,
    ]
    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_CHECKERS, _features.flake8)
@_invoke.task("clean.py")
def flake8(ctx):
    """Run flake8"""
    if ctx.get("package"):
        if _os.path.exists(ctx.shell.native(ctx.package + ".py")):
            candidate = ctx.package + ".py"
        else:
            candidate = ctx.package
        cmd = [ctx.which("flake8"), candidate]
        with ctx.shell.root_dir():
            ctx.run(ctx.c(cmd), echo=True)

    cmd = [ctx.which("flake8"), "tasks"]
    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_CHECKERS, _features.black)
@_invoke.task("clean.py")
def black(ctx):
    """Run black"""
    cmd = [ctx.which("black")] + ctx.s(
        "-q --check --diff --color --config pyproject.toml ."
    )

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_CHECKERS, _features.isort)
@_invoke.task("clean.py")
def isort(ctx):
    """Run isort"""
    cmd = [ctx.which("isort")]
    cmd += ctx.s("--check --diff --color --settings-path pyproject.toml")
    cmd += ["tasks"]
    if ctx.get("package"):
        cmd += [ctx.package]
    if _os.path.exists(ctx.shell.native("tests")):
        cmd += ["tests"]

    with ctx.shell.root_dir():
        ctx.run(ctx.c(cmd), echo=True)


@_tasks.optional(_CHECKERS_TF, _features.terraform)
@_invoke.task()
def tf_fmt(ctx):
    """Run terraform fmt -check"""
    base_cmd = [ctx.which("terraform"), "-chdir=" + ctx.paths.terraform]
    with ctx.shell.root_dir():
        ctx.run(ctx.c(base_cmd + ctx.s("fmt -recursive -check")), echo=True)


@_tasks.optional(_CHECKERS_TF, _features.terraform)
@_invoke.task()
def tf_validate(ctx):
    """Run terraform validate"""
    base_cmd = [ctx.which("terraform"), "-chdir=" + ctx.paths.terraform]
    with ctx.shell.root_dir():
        ctx.run(
            ctx.c(base_cmd + ctx.s("init -backend=false")),
            hide=True,
            echo=True,
        )
        ctx.run(ctx.c(base_cmd + ctx.s("validate")), echo=True)


# @_tasks.optional(_CHECKERS_SAM, _features.sam)
# @_invoke.task()
# def sam_validate(ctx):
#     """Run sam validate"""
#     cmd = (
#         [ctx.which("sam")]
#         + ctx.s("validate --lint -t")
#         + [ctx.cloudformation.template]
#     )
#     with ctx.shell.root_dir():
#         ctx.run(ctx.c(cmd), echo=True)


# Collections of checks
# =====================


@_tasks.optional(_CHECKERS_TF)
@_invoke.task(name="tf")
def terraform(ctx):
    """Run terraform checks"""
    _tasks.execute(ctx, *["check." + checker for checker in _CHECKERS_TF])


@_tasks.optional(_CHECKERS_SAM)
@_invoke.task()
def sam(ctx):
    """Run sam checks"""
    _tasks.execute(ctx, *["check." + checker for checker in _CHECKERS_SAM])


_CHECKERS += _CHECKERS_TF
_CHECKERS += _CHECKERS_SAM


@_tasks.optional(_CHECKERS)
@_invoke.task(name="all", default=True)
def all_(ctx):
    """Run all checkers"""
    _tasks.execute(ctx, *["check." + checker for checker in _CHECKERS])



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Build Tasks
~~~~~~~~~~~

"""

import invoke as _invoke

from . import _dist
from . import _release
from . import _version

NAMESPACE = "build"


@_invoke.task("doc.doc")
def dist(ctx):
    """Build distribution"""
    fakeroot = ctx.shell.frompath("fakeroot")
    with ctx.shell.root_dir():
        ctx.shell.rm_rf("build", "dist")
        ctx.run(
            ctx.c("%s python setup.py sdist --formats tar,zip", fakeroot),
            echo=True,
        )

        files = list(ctx.shell.files("dist", "*.zip"))
        digestname = files[0][:-3] + "digests"
        for name in ctx.shell.files("dist", "*.tar"):
            files.append(_dist.compress(ctx, name, "gzip", ".gz"))
            files.append(_dist.compress(ctx, name, "bzip2", ".bz2"))
            files.append(_dist.compress(ctx, name, "xz", ".xz"))
            ctx.shell.rm(name)
        files = [name for name in files if name]

        _dist.digest(ctx, files, digestname)
        _dist.copy_changes(ctx)


@_invoke.task()
def version(ctx):
    """Version"""
    with ctx.shell.root_dir():
        _version.update(ctx)


@_invoke.task("doc.doc")
def release(ctx):
    """Release"""
    with ctx.shell.root_dir():
        _release.check_committed(ctx)
        version(ctx)
        _release.add_tag(ctx)

    # _doc.doc(ctx)  # dependencies are not called, hence put as regular
    # dep here
    dist(ctx)



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Compile tasks
~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from . import _features
from . import pypi as _pypi
from ._inv import tasks as _tasks


@_tasks.optional(None, _features.python_package)
@_invoke.task(default=True)
def compile(ctx):  # pylint: disable = redefined-builtin
    """Compile the package"""
    with ctx.shell.root_dir():
        ctx.run(
            ctx.c("pip install -i %s -e .", _pypi.index_url(ctx)),
            env=dict(_os.environ, SETUP_CEXT_REQUIRED="1"),
        )



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Task management
~~~~~~~~~~~~~~~

"""
__all__ = ["by_name", "execute", "exists", "new_context"]

import collections as _collections
import importlib as _importlib
import os as _os
import re as _re

import invoke as _invoke

from . import shell as _shell
from . import util as _util

_TASKS = None
_SETTINGS = None


def exists(name):
    """
    Check if a task exists

    Parameters:
      name (str):
        The fully qualified task name

    Returns:
      bool: Does the task exist?
    """
    assert _TASKS is not None

    return name in _TASKS


def by_name(name):
    """
    Find task by name

    Parameters:
      name (str):
        The fully qualified task name

    Returns:
      invoke.Task: The task

    Raises:
      KeyError: Task not found
    """
    assert _TASKS is not None

    return _TASKS[name]


def execute(*args, **kwargs):  # pylint: disable = invalid-name
    """
    execute(ctx, name, **kwargs)

    Execute a task like invoke does (including pre + post).

    All arguments have to be given via kwargs.

    Parameters:
      *args:
        List of (task-name, kwargs) tuples. If there's no kwargs for a task,
        the task-name only is sufficient. If there's only one task-name given,
        kwargs is applied to that task. Otherwise kwargs must be empty.

      **kwargs:
        Arguments if a single task is given

    Returns:
      dict: A dict mapping task objects to their return values (as returned
            from ``invoke.Exceutor().execute()``)
    """
    assert _TASKS is not None

    ctx, tasks = args[0], list(args[1:])
    if len(tasks) == 1 and isinstance(tasks[0], _util.basestring_):
        tasks = [(tasks[0], kwargs)]
    elif kwargs:
        raise AssertionError("Multiple tasks cannot share the same kwargs")

    for idx, task in enumerate(tasks):
        if isinstance(task, _util.basestring_):
            tasks[idx] = (task, {})

    return _invoke.Executor(_TASKS, ctx.config).execute(*tasks)


def new_context():
    """
    Create a new context

    Typically only needed if a function is run "independent" from invoke.

    Returns:
      invoke.Context: The new context
    """
    assert _TASKS is not None

    config = _invoke.Config()
    collection_config = _TASKS.configuration()
    config.load_collection(collection_config)
    config.load_shell_env()
    return _invoke.context.Context(config)


def optional(collection, featureflag=None):
    """
    Decorator for optional task

    Parameters:
      collection (list):
        Collection to append the task name to if the task is enabled. Can be
        ``None`` if no collection is desired.

      featureflag (callable):
        Feature flag function. If it's not None, it's called without
        arguments. The return value is treated as a bool. If true, the task
        will be returned by the decorator and the task name will be added to
        the collection. Otherwise ``None`` will be returned.

        If featureflag is ``None``, the collection will be checked as a bool.
        If true the task will be returned, ``None`` otherwise.

    Returns:
      callable: inner function accepting the task and taking the decision.
    """

    def inner(func):
        """
        Inner proxy, taking the return decision

        Parameters:
          func (Task):
            The optional task

        Returns:
          callable: The task or ``None``
        """
        if featureflag is None:
            if collection:
                return func

        elif featureflag(_SETTINGS):
            if collection is not None:
                collection.append(func.name)
            return func

        return None

    return inner


def setup_tasks(env):
    """
    Find available tasks (cached)

    Parameters:
      env (dict-like):
        The settings

    Returns:
      invoke.Collection: The tasks
    """
    global _SETTINGS, _TASKS  # pylint: disable = global-statement
    if _TASKS is not None:
        return _TASKS
    _SETTINGS = env

    path = _os.path.dirname(_os.path.dirname(_os.path.abspath(__file__)))
    pkg = _os.path.basename(path)
    torepair, space = [], _collections.defaultdict(dict)

    for module in _find_modules():
        name = getattr(module, "NAMESPACE", None)
        if not name:
            name = module.__name__[len(pkg) + 1 :]  # noqa

        for item in vars(module).values():
            if not isinstance(item, _invoke.Task):
                continue

            torepair.append((name, item))
            space[name][item.name] = item

    result = _invoke.Collection(
        **{
            name: _invoke.Collection(*tasks.values())
            for name, tasks in space.items()
        }
    )
    result.configure(env)

    # pre and post can be passed as string - map them here
    for name, task in torepair:
        for attr in ("pre", "post"):
            res = []
            for dep in getattr(task, attr):
                if not isinstance(dep, _util.basestring_):
                    res.append(dep)
                    continue

                if "." in dep:
                    depspace, dep = dep.rsplit(".", 1)
                else:
                    depspace = name
                res.append(space[depspace][dep])
            setattr(task, attr, res)

    _TASKS = result
    return result


def _find_modules():
    """
    Find modules with tasks

    Returns:
      iterable: Modules found
    """
    path, prune = _os.path.split(_os.path.dirname(_os.path.abspath(__file__)))
    pkg = _os.path.basename(path)
    pkg_path = pkg + "/"

    is_valid_file = _re.compile(r"^[a-zA-Z_]+[a-zA-Z0-9_]*\.py$").match
    is_valid_dir = _re.compile(r"^[a-zA-Z_]+[a-zA-Z0-9_]*$").match
    init = "__init__.py"

    for filename in _shell.files(path, "[!.]*.py", prune=[prune]):
        assert filename.startswith(pkg_path)
        filename = filename[len(pkg_path) :]  # noqa
        dirname, basename = _os.path.split(filename)
        odirname = dirname

        # Skip root __init__.py or non-id names
        if filename == init or not is_valid_file(basename):
            continue

        # Check all sub-path parts if any
        while dirname:
            if not is_valid_dir(_os.path.basename(dirname)):
                break
            if not _os.path.exists(_os.path.join(path, dirname, init)):
                break
            dirname = _os.path.dirname(dirname)

        # import module/package and yield if there are any tasks
        else:
            modname = filename[:-3] if basename != init else odirname
            modname = ".".join(_shell.pathparts(modname))
            qname = "%s.%s" % (pkg, modname)

            module = _importlib.import_module(qname)
            for item in vars(module).values():
                if isinstance(item, _invoke.Task):
                    yield module
                    break



# -*- coding: ascii -*-
#
# Copyright 2019 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Meta information
~~~~~~~~~~~~~~~~

"""

from .._inv import shell as _shell

# pylint: disable = import-outside-toplevel


def find_meta():
    """
    Find package meta data

    Returns:
      dict: Package metadata
    """
    # pylint: disable = import-error, no-name-in-module
    from build import util as _build_util

    if find_meta.found is None:
        find_meta.found = _build_util.project_wheel_metadata(
            _shell.native(".")
        )

    return find_meta.found


find_meta.found = None



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Build Tasks
~~~~~~~~~~~

"""
from __future__ import print_function

import itertools as _it
import os as _os
import platform as _platform
import re as _re

import invoke as _invoke

from .. import _features
from .._inv import tasks as _tasks
from . import _container

NAMESPACE = "build"


@_tasks.optional(None, _features.python_wheels)
@_invoke.task()
def wheels(ctx, arches=None, pythons=None):
    """
    Build wheels

    Parameters:
      arches (str):
        Space separated list of architectures. If not supplied it defaults to
        the current machine type. Except if the current machine type is
        "x86_64", then if defaults to "x86_64 i686 aarch64" (using
        binfmt_misc in combination with multiarch/qemu-user-static for the
        latter).

      pythons (str):
        Space separated list of pythons (36 37 etc). If omitted, all possible
        python versions will be built. Only relevant for binary builds.
    """
    if ctx.wheels.build == "binary":
        return _build_binary(ctx, arches=arches, pythons=pythons)

    assert not arches
    assert ctx.wheels.build in ("universal", "simple")
    return _build_universal(ctx)


def _build_universal(ctx):
    """Build universal / simple wheel"""
    path = ".bdist"

    with ctx.shell.root_dir():
        ctx.shell.rm_rf(path)

        cmd = ctx.s("pip wheel --no-binary :all: --no-cache --no-deps -w")
        cmd += [path]

        for package in ctx.shell.files(
            "dist/", "%s[-_]*.tar.gz" % (ctx.package,)
        ):
            ctx.run(ctx.c(cmd + [package]), echo=True)
            _finalize_bdist(ctx, path)


def _build_binary(ctx, arches=None, pythons=None):
    """
    Build binary wheels

    Parameters:
      arches (str):
        Space separated list of architectures. If not supplied it defaults to
        the current machine type. Except if the current machine type is
        "x86_64", then if defaults to "x86_64 i686 aarch64" (using
        binfmt_misc in combination with multiarch/qemu-user-static for the
        latter).

      pythons (str):
        Space separated list of pythons (36 37 etc). If omitted, all possible
        python versions will be built. Only relevant for binary builds.
    """
    path = ".bdist"
    arches = _ensure_arches(ctx, arches)
    build_sh = _os.path.join(_os.path.dirname(__file__), "wheel.sh")
    wanted_pythons = set(pythons.split()) if pythons else None

    with ctx.shell.root_dir():
        ctx.shell.rm_rf(path)
        ctx.shell.mkdir_p(path)

        for package in ctx.shell.files(
            "dist/", "%s-*.tar.gz" % (ctx.package,)
        ):
            for image, cmd, libc, py in _pypa_containers(ctx, arches):
                pythons = (
                    py
                    if wanted_pythons is None
                    else (set(py) & wanted_pythons)
                )
                if not pythons:
                    continue

                name = "wheel-%s" % (_os.path.basename(package),)
                cmd += [
                    "/io/build.sh",
                    libc,
                    "/io/pkg.tar.gz",
                    " ".join(sorted(pythons)),
                ]

                with _container.Container(ctx, name, image) as container:
                    container.execute(ctx.s("mkdir /io"))
                    container.cp_in(package, "/io/pkg.tar.gz")
                    container.cp_in(build_sh, "/io/build.sh")
                    container.execute(ctx.s("chmod +x /io/build.sh"))
                    container.execute(cmd)
                    container.cp_out("/io/dist/.", path)

                _finalize_bdist(ctx, path)


def _finalize_bdist(ctx, path):
    """
    Finalize wheel names, move files to dist/

    Parameters:
      path (str):
        wheel path
    """
    # collapse file names
    multi_sub = _re.compile(r"(?:[.-]manylinux[^.]+)+").sub
    tomove = []
    for name in _os.listdir(path):
        if not name.endswith(".whl"):
            continue
        if "manylinux" not in name:
            continue

        pick = _best_manylinux(name)
        if pick:
            tomove.append(
                (name, multi_sub((lambda _, p=pick: "-" + p), name))
            )
    for old, new in tomove:
        if old == new:
            continue
        _os.rename(_os.path.join(path, old), _os.path.join(path, new))

    if path == "dist":
        return

    # Move to dist/
    for name in _os.listdir(path):
        old = "%s/%s" % (path, name)
        new = "dist/%s" % (name,)
        print("%s -> %s" % (old, new))
        _os.rename(ctx.shell.native(old), ctx.shell.native(new))


def _pypa_containers(ctx, arches):
    """
    Create a list of containers to execute

    Parameters:
      arches (list):
        List of arches to apply

    Yields:
      tuple: image, command prefix, libc, python list
    """
    for arch in arches:
        spec = ctx.wheels.specs[arch]
        prefix = "linux32" if arch == "i686" else ""
        for libc in ("manylinux", "musllinux"):
            groups = sorted(
                (
                    (tup[0], tup[1][libc])
                    for tup in spec.items()
                    if libc in tup[1]
                ),
                key=lambda x, libc=libc: (x[1], x[0]),
            )
            for v, group in _it.groupby(groups, key=lambda x: x[1]):
                if "_" in v:
                    v = "_" + v
                if ":" in v:
                    v, tag = v.split(":", 1)
                else:
                    tag = "latest"

                yield (
                    "quay.io/pypa/%s%s_%s:%s" % (libc, v, arch, tag),
                    [prefix] if prefix else [],
                    libc,
                    sorted(item[0] for item in group),
                )


def _ensure_arches(ctx, arches):
    """
    Install multiarch qemu

    Parameters:
      arches (str):
        The desired architectures

    Returns:
      list: The split architecture strings
    """
    machine = _platform.machine()
    system = _platform.system()
    if arches is None:
        if machine == "x86_64":
            arches = "x86_64 i686 aarch64"
        else:
            arches = machine
    arches = arches.split()

    if system == "Linux":
        for arch in arches:
            if machine == arch:
                continue
            if machine == "x86_64":
                if arch == "i686":
                    continue

                ctx.run(
                    ctx.c(
                        _container.DockerCommand(ctx).run(
                            "multiarch/qemu-user-static",
                            command=["--reset", "-p", "yes"],
                            remove=True,
                            privileged=True,
                        )
                    ),
                    echo=True,
                )

                break

    return arches


def _best_manylinux(name):
    """
    Find best manylinux variant produced by auditwheel

    Parameters:
      name (str):
        filename (basename) of the package

    Returns:
      str: The best version, or ``None`` if it couldn't be determined
    """
    tag_list_search = _re.compile(r"(?:[.-]manylinux[^.]+)+").search

    tags = tag_list_search(name)
    if not tags:
        return None
    sortable = []
    for tag in tags.group(0).strip("-.").split("."):
        assert tag.startswith("manylinux")
        ver = tag[len("manylinux") :]
        if ver.startswith(("1_", "2010_", "2014_")):
            sortable.append(((1, int(ver.split("_")[0])), tag))
            continue
        assert ver.startswith("_")
        ver = tuple(map(int, ver[1:].split("_")[:2]))
        assert ver[0] > 1
        sortable.append((ver, tag))

    if not sortable:
        return None

    return min(sortable)[-1]



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared code for dependency files
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import re as _re

from .. import _parse

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.suggest.file_model")


class Model(object):
    """
    File model base

    Attributes:
      parts (list):
        The file parts (strings vs literals)
    """

    #: The file type
    #:
    #: :Type: str
    type = None

    #: Encoding of the file
    #:
    #: :Type: str
    encoding = "latin-1"

    #: Require a version spec when parsing the file?
    #:
    #: :Type: bool
    require_version = None

    def __init__(self, parts):
        """
        Initialization

        Parameters:
          parts (list):
            The file parts
        """
        self.parts = parts

    @classmethod
    def by_name(cls, filename):
        """
        Construct by filename

        Parameters:
          filename (str):
            The filename

        Returns:
          Model: New container instance
        """
        with open(filename, "rb") as fp:
            return cls(cls._parse_stream(fp))

    @classmethod
    def _parse_stream(cls, fp):
        """
        Parse from binary file stream

        Parameters:
          fp (file):
            Binary file stream

        Returns:
          list: The parts
        """
        raise NotImplementedError()


class ReqFile(Model):
    """Requirement file container"""

    type = "text"
    encoding = "latin-1"
    require_version = False

    @classmethod
    def _parse_stream(cls, fp):
        """
        Parse req file
        """
        parts = []
        for line in fp:
            line = line.decode(cls.encoding)
            xline = line.strip()
            if not xline or xline.startswith("#"):
                parts.append(LiteralPart(line))
            else:
                parts.append(StringPart.from_text(line))

        return parts


class SetupFile(Model):
    """Setup file container"""

    type = "setup"
    encoding = "latin-1"
    require_version = True

    @classmethod
    def _parse_stream(cls, fp):
        """
        Parse setup file
        """
        parts = []
        file_content = fp.read().decode(cls.encoding)

        last = 0
        for match in _parse.find_py_str_or_comment(file_content):
            # comments are just skipped. We look for them anyway, because they
            # might contain valid string patterns (which we will ignore)
            if match.group(2):
                continue

            if match.start() > last:
                parts.append(LiteralPart(file_content[last : match.start()]))
            last = match.end()
            raw = match.start() > 0 and file_content[match.start() - 1] == "r"
            parts.append(StringPart.from_python(raw, match.group(1)))

        if len(file_content) > last:
            parts.append(LiteralPart(file_content[last:]))

        return parts


class TomlFile(Model):
    """
    TOML file container
    """

    type = "toml"
    encoding = "utf-8"
    require_version = True

    @classmethod
    def _parse_stream(cls, fp):
        """
        Parse toml file
        """
        parts = []
        file_content = fp.read().decode(cls.encoding)

        last = 0
        for match in _parse.find_toml_str_or_comment(file_content):
            # comments are just skipped. We look for them anyway, because they
            # might contain valid string patterns (which we will ignore)
            if match.group(2):
                continue

            if match.start() > last:
                parts.append(LiteralPart(file_content[last : match.start()]))
            last = match.end()
            raw = file_content[match.start()] == "'"
            parts.append(StringPart.from_toml(raw, match.group(1)))

        if len(file_content) > last:
            parts.append(LiteralPart(file_content[last:]))

        return parts


class StringPart(object):
    """
    String part (of a dep file)

    Attributes:
      raw (bool):
        Is it a raw string?

      quote (str):
        The quote used

      content (str):
        The string content, escaped quotes stripped if not raw
    """

    #: Is it a literal part?
    #:
    #: :Type: bool
    is_literal = False

    def __init__(self, raw, quote, content):
        """
        Initialization

        :Parameters:
          raw (bool):
            Is it a raw string?

          quote (str):
            The quote used

          content (str):
            The string content, escaped quotes stripped if not raw
        """
        self.raw = raw
        self.quote = quote
        self.content = content

    def __eq__(self, other):
        """
        Return equality

        Returns:
          bool: Is it equal to other?
        """
        if not isinstance(other, self.__class__):
            return False
        return (self.raw, self.quote, self.content) == (
            other.raw,
            other.quote,
            other.content,
        )

    def __ne__(self, other):
        """
        Return non-equality

        Returns:
          bool: Is it not equal to other?
        """
        return not self == other

    def __repr__(self):
        """
        Create repr of this container

        Returns:
          str: Representation of the container
        """
        return "%s(%r, %r, %r)" % (
            self.__class__.__name__,
            self.raw,
            self.quote,
            self.content,
        )

    @classmethod
    def from_python(cls, raw, content):
        """
        Construct string from python-encoded

        Parameters:
          raw (bool):
            Is it a raw string?

          content (str):
            The string content from python, including quotes

        Returns:
          String: New string container
        """
        if content.startswith(('"""', "'''")):
            quote, content = content[:3], content[3:-3]
        else:
            quote, content = content[0], content[1:-1]
        if not raw:
            content = _re.sub(
                r"(?s)\\(.)",
                lambda m: m.group(m.group(1) in "'\""),
                content,
            )
        return cls(raw, quote, content)

    from_toml = from_python

    @classmethod
    def from_text(cls, content):
        """
        Construct string from text

        Parameters:
          content (str):
            The string content from text file

        Returns:
          String: New string container
        """
        return cls(True, "", content)

    def __str__(self):
        """
        Rebuild the string part as close to the original as possible

        Returns:
          str: The string, quoted, quotes escaped if needed
        """
        content = self.content
        if not self.raw:
            content = content.replace(self.quote, "\\%s" % (self.quote,))
        return "%s%s%s" % (self.quote, content, self.quote)


class LiteralPart(object):
    """
    Literal part (of a dep file)

    Attributes:
      content (str):
        The part content
    """

    #: Is it a literal part?
    #:
    #: :Type: bool
    is_literal = True

    def __init__(self, content):
        """
        Initialization

        Parameters:
          content (str):
            The part content
        """
        self.content = content

    def __eq__(self, other):
        """
        Return equality

        Returns:
          bool: Is it equal to other?
        """
        if not isinstance(other, self.__class__):
            return False
        return self.content == other.content

    def __ne__(self, other):
        """
        Return non-equality

        Returns:
          bool: Is it not equal to other?
        """
        return not self == other

    def __repr__(self):
        """
        Create repr of this container

        Returns:
          str: Representation of the container
        """
        return "%s(%r)" % (self.__class__.__name__, self.content)

    def __str__(self):
        """
        Rebuild the part as string

        Returns:
          str: The part as string
        """
        return self.content



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Dist building code
~~~~~~~~~~~~~~~~~~

"""

import os as _os

# pylint: disable = import-outside-toplevel


def compress(ctx, filename, cmd, ext):
    """Compress tar file"""
    prog = ctx.shell.frompath(cmd)
    if prog is None:
        return None

    ctx.run(
        "%s -c9 <%s >%s"
        % (ctx.q(prog), ctx.q(filename), ctx.q(filename + ext)),
        echo=True,
    )
    return filename + ext


def digest(ctx, files, digestname):
    """Make digest file"""
    import hashlib as _hashlib

    result = {}
    for filename in files:
        basename = _os.path.basename(filename)
        result[basename] = []
        for method in ("md5", "sha1", "sha256"):
            sign = getattr(_hashlib, method)()
            with open(filename, "rb") as fp:
                for block in iter(lambda fp=fp: fp.read(8192), b""):
                    sign.update(block)
            result[basename].append(sign.hexdigest())

    with open(digestname, "wb") as fp:
        fp.write(b"\n# The file contains MD5, SHA1 and SHA256 digests\n")
        fp.write(b"# Check archive integrity with, e.g. md5sum -c\n")
        fp.write(b"# Check digest file integrity with PGP\n\n")

        for basename, digests in sorted(result.items()):
            for sign in digests:
                fp.write(("%s *%s\n" % (sign, basename)).encode("utf-8"))

    _sign_inline(ctx, digestname)


def _sign_inline(ctx, filename):
    """Sign file (inline)"""
    gpg = ctx.shell.frompath("gpg")
    ctx.run(
        ctx.c(
            """
        %s --output %s --clearsign -- %s
    """,
            gpg,
            filename + ".signed",
            filename,
        ),
        echo=True,
    )
    _os.rename(filename + ".signed", filename)


def copy_changes(ctx):
    """Copy changelog"""
    version = ctx.run("python setup.py --version", hide=True).stdout.strip()
    from_ = "CHANGES"
    to_ = "dist/CHANGES-%s" % (version,)
    ctx.shell.cp(from_, to_)



# -*- coding: ascii -*-
#
# Copyright 2007 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Build Tasks
~~~~~~~~~~~

"""

import contextlib as _contextlib
import os as _os
import tempfile as _tempfile

from ..util import package as _package


@_contextlib.contextmanager
def sdist(ctx, name):
    """
    Create temporary package directory as a context

    Parameters:
      name (str):
        The sub-package name

    Returns:
      The directory name
    """
    dotsdist = ".sdist"
    try:
        with ctx.shell.root_dir():
            ctx.shell.mkdir_p(dotsdist)
            directory = _tempfile.mkdtemp(dir=dotsdist)

            # cp target cannot exist before
            target = _os.path.join(directory, "x")
            ctx.shell.cp_r(".", target, ignore=_ignore(ctx))

            ext_map = {
                ".py": "setup.py",
                ".toml": "pyproject.toml",
                ".MANIFEST": "MANIFEST.in",
            }
            for ext, dest in ext_map.items():
                dest = _os.path.join(target, dest)
                ctx.shell.rm(dest)

                src = _os.path.join(target, "setups", name + ext)
                if _os.path.isfile(src):
                    with open(src, "rb") as fp:
                        content = fp.read().decode("latin-1")
                    if "$BASE" in content or "$VERSION" in content:
                        meta = _package.find_meta()
                        content = content.replace(
                            "$VERSION", meta["Version"]
                        ).replace("$BASE", meta["Name"])
                        with open(dest, "wb") as fp:
                            fp.write(content.encode("latin-1"))
                    else:
                        ctx.shell.cp(src, dest)
            if not _os.path.isfile(_os.path.join(target, "MANIFEST.in")):
                with open(_os.path.join(target, "MANIFEST.in"), "wb"):
                    pass
            ctx.shell.rm_rf(_os.path.join(target, "setups"))

            _os.chdir(target)
            yield target
    finally:
        ctx.shell.rm_rf(dotsdist)


def _ignore(ctx):
    """
    Make ignore function (as expected by shutil.copytree)

    Returns:
      callable: The copytree ignore callback function
    """
    # Config option to add more files/directories to ignore
    extra = ctx.get("setup_ignore")

    # Collect all checked-in files, those are good.
    with ctx.shell.root_dir():
        files = set(
            _os.path.normpath(item)
            for item in ctx.run("git ls-files -z", hide=True).stdout.split(
                "\0"
            )
        )
    for item in list(files):
        dirname = _os.path.dirname(item)
        if dirname:
            files.add(dirname)

    root = []
    extra = set(_os.path.normpath(item) for item in extra or ())

    def ignorable(node, children):
        """
        Ignore function as called by shutil.copytree

        Parameters:
          node (str):
            The directory

          children (list):
            The directory entries

        Returns:
          list: The entries to ignore
        """
        if not root:
            root.append(node)
        node = node[len(root[0]) + 1 :]

        # Collect entries to ignore
        result = []
        for item in children:
            path = _os.path.normpath(
                _os.path.join(node, item) if node else item
            )

            # explicit ignores takes precedence
            if path in extra:
                result.append(item)

            # If it looks weird, check back with the git list
            elif item.startswith(".") or item.endswith(".egg-info"):
                if path not in files:
                    result.append(item)

        return result

    return ignorable



# -*- coding: ascii -*-
#
# Copyright 2024 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities
~~~~~~~~~

"""

import functools as _ft

basestring_ = (
    str
    if str is not bytes
    else basestring  # noqa pylint: disable = undefined-variable
)


def cached(func):
    """
    Caching decorator

    Parameters:
      func (callable):
        The function to cache the result of

    Returns:
        The decorated function
    """
    cache = getattr(_ft, "cache", None)
    if cache is not None:
        return cache(func)

    store = {}

    @_ft.wraps(func)
    def inner(*args, **kwargs):
        """The cache wrapper"""
        key = tuple(list(args) + sorted(kwargs.items()))
        try:
            return store[key]
        except KeyError:
            store[key] = result = func(*args, **kwargs)
            return result

    return inner


def dictmerge(dict1, dict2, resolve_left=None):
    """
    Merge nested dict2 into dict1

    Parameters:
      dict1 (dict or adict):
        Dict to update

      dict2 (dict or adict):
        Dict to merge

      resolve_left (callable):
        A resolver to be called on each callable left-side item. If omitted or
        ``None``, the items will be passed through as-is

    Returns:
      dict or adict: The updated dict1. If it was an adict, a new adict is
                     returned.
    """
    type1 = type(dict1)
    if isinstance(dict2, adict):
        type1 = adict
    elif type1 not in (dict, adict):
        type1 = adict

    if dict1 is None:
        if not isinstance(dict2, (dict, adict)):
            return dict2
        dict1 = {}

    dict1 = dict(dict1.items())
    if resolve_left is not None:
        for key, value in list(dict1.items()):
            if callable(value) and isinstance(dict2.get(key), (adict, dict)):
                dict1[key] = resolve_left(value)

    for key, value in dict2.items():
        if isinstance(value, (dict, adict)):
            value = dictmerge(
                dict1.get(key, {}),
                value,
                resolve_left=resolve_left,
            )
        dict1[key] = value

    return type1(dict1)


class adict(object):
    """attribute dict"""

    # pylint: disable = invalid-name, missing-docstring

    def __init__(self, *args, **kwargs):
        """
        Initialization

        Parameters:
          *args:
            Positional dict args

          **kwargs:
            Keyword dict args
        """
        self.__x__ = dict(*args, **kwargs)

    def __repr__(self):
        """
        Return debug representation

        Returns:
          str: Debug representation
        """
        return "%s(%r)" % (self.__class__.__name__, self.__x__)

    def __bool__(self):
        """
        Return bool value of the object

        Returns:
          bool: False if empty, True otherwise
        """
        return bool(self.__x__)

    __nonzero__ = __bool__

    def __iter__(self):
        """
        Return key iterator

        Returns:
          iterable: New key iterator
        """
        return iter(self.__x__)

    def __getitem__(self, name):
        """
        Get value by subscript

        Parameters:
          name (str):
            The key

        Returns:
          any: The value

        Raises:
          KeyError: Key not found
        """
        return self.__x__[name]

    def __setitem__(self, name, value):
        """
        Set value by subscript

        Parameters:
          name (str):
            The key

          value (any):
            The value
        """
        self.__x__[name] = value

    def __getattr__(self, name):
        """
        Get value for dot notation

        Parameters:
          name (str):
            The key

        Returns:
          any: The value

        Raises:
          AttributeError: Key not found
        """
        if name == "__setstate__":
            raise AttributeError(name)
        try:
            return super(adict, self).__getattr__(name)
        except AttributeError:
            try:
                return self.__x__[name]
            except KeyError:
                raise AttributeError(  # pylint: disable = raise-missing-from
                    name
                )

    def __setattr__(self, name, value):
        """
        Set value with dot notation

        Parameters:
          name (str):
            The key

          value (any):
            The value
        """
        if name == "__x__" or hasattr(self.__class__, name):
            super(adict, self).__setattr__(name, value)
        else:
            self.__x__[name] = value

    def keys(self):
        """
        Create key iterator

        Returns:
          iterable: New key iterator
        """
        return self.__x__.keys()

    def items(self):
        """
        Create key/value iterator

        Returns:
          iterable: New key/value iterator
        """
        return self.__x__.items()

    def get(self, *args, **kwargs):
        """
        Get value with default

        Parameters:
          *args:
            Positional dict.get args

          **kwargs:
            Keyword dict.get args

        Returns:
          any: The value or the default (if the key doesn't exist)
        """
        return self.__x__.get(*args, **kwargs)

    def pop(self, *args, **kwargs):
        """
        Pop value

        Parameters:
          *args:
            Positional dict.pop args

          **kwargs:
            Keyword dict.pop args

        Returns:
          any: The value or the default (if the key doesn't exist)
        """
        return self.__x__.pop(*args, **kwargs)



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Inspect current dependencies
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import collections as _collections
import json as _json
import logging as _logging
import os as _os
import re as _re

from .. import _parse
from . import _index
from . import _req_file

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.inspect")


def inspect_dependencies(ctx, config, verbose=True, debug=False):
    """
    Inspect dependencies

    Parameters:
      config (adict):
        deps config

      verbose (bool):
        Verbose mode? Default: true

      debug (bool):
        Debug mode? Default: false

    Returns:
      dict: package -> info mapping
    """
    logger.setLevel("DEBUG" if debug else ("INFO" if verbose else "WARN"))

    requirements, _ = _scan_toplevel(ctx, config.toplevel)
    index = _index.Index()

    return {
        name: _inspect_package(index, name, version, places)
        for (name, version), places in requirements.items()
    }


def _scan_toplevel(ctx, req_files):
    """
    Find toplevel packages

    Parameters:
      req_files (list):
        Requirement file names

    Returns:
      tuple: {(normname, installed-version) -> [{line, requirement}]},
             [constraint]
    """
    installed = _find_installed(ctx)

    req = _collections.defaultdict(list)
    con = []

    for item in _req_file.tokens(req_files, base=ctx.shell.root):
        if item.type == _req_file.REQUIREMENT:
            name = _parse.normalize(item.value.name)
            version = installed.get(name, {}).get("version")
            key = (name, version)
            logger.debug("Parsed %r -> %r", item.line.content.strip(), key)
            req[key].append(dict(line=item.line, requirement=item.value))
            continue

        if item.type == _req_file.EDITABLE:
            for parsed, extras in _find_editable_deps(ctx, installed, item):
                name = _parse.normalize(parsed.name)
                version = installed.get(name, {}).get("version")
                key = (name, version)
                logger.debug("Found editable dependency %r", key)
                req[key].append(
                    dict(line=None, requirement=parsed, extras=extras)
                )
            continue

        if item.type == _req_file.CONSTRAINT:
            con.append(item.value)
            continue

    return dict(req), con


def _inspect_package(index, name, version, places):
    """
    Inspect a single package (in multiple locations)

    Parameters:
      index (PackageIndex):
        package index connector

      wset (WorkingSet):
        Working set to use

      name (str):
        The (normalized) package name

      version (str):
        Installed version (or ``None``)

      places (list):
        Places where the requirement was found
    """
    logger.info("Inspecting %s (%s)", name, version)

    scanner = index.scanner(version)
    locations = [
        _inspect_place(scanner, place)
        for place in places
        if _place_evaluates(place)
    ]
    return dict(
        installed=version,
        locations=locations,
        updates=scanner.result() or None,
    )


def _inspect_place(scanner, place):
    """
    Inspect a single place

    Parameters:
      index (PackageIndex):
        package index connector

      wset (WorkingSet):
        Working set to use

      place (dict):
        The place to inspect

      version (Version):
        parsed installed version

      version_info (dict):
        Index version info. Modified in-place.

    Returns:
      dict: The fully inspected place description
    """
    result = dict(requirement=place["requirement"])
    if place["line"]:
        result["file"] = dict(
            name=place["line"].file,
            line=place["line"].lineno,
            content=place["line"].content,
        )
        logger.debug("Adding location %r", result["file"])

    scanner.inspect(place["requirement"])
    return result


def _place_evaluates(place):
    """
    Does a place evaluate in the current environment?

    Parameters:
      place (dict):
        The place to inspect

    Returns:
      bool: The the place's requirement evaluate in the current environment?
    """
    if not place["requirement"].marker:
        return True

    if not place.get("extras"):
        if place["requirement"].marker.evaluate({}):
            return True
    else:
        for extra in place["extras"]:
            env = dict(extra=_parse.safe_extra(extra))
            if place["requirement"].marker.evaluate(env):
                return True

    logger.debug(
        "Ignoring %r because it doesn't evaluate", str(place["requirement"])
    )
    return False


def _find_editable_deps(ctx, installed, spec):
    """
    Find direct dependencies for editable installation

    Parameters:
      installed (dict):
        Installed packages

      spec (Token):
        The editable dep file token

    Returns:
      iterable: List of (requirement, needed-extras) tuples
    """
    pkg_name = _find_editable(ctx, installed, spec)
    if pkg_name:
        parts = spec.value.split("[", 1)
        parts[0] = pkg_name
        extras = sorted(_parse.requirement("[".join(parts)).extras)
        logger.debug("Extras for editable installation: %r", extras)

        try:
            try:
                from importlib import metadata as _metadata
            except ImportError:
                import importlib_metadata as _metadata
        except ImportError:
            import pkg_resources as _pkg_resources

            dist = _pkg_resources.working_set.by_key[pkg_name]
            for item in dist.requires(extras):
                yield item, extras
        else:
            for item in _metadata.requires(pkg_name) or ():
                parsed = _parse.requirement(item)
                yield parsed, extras


def _find_editable(ctx, installed, spec):
    """
    Find editable package installation

    Parameters:
      installed (dict):
        Installed packages

      spec (Token):
        The editable dep file token

    Returns:
      str: The normalized name
    """
    local_path = spec.value.split("[", 1)[0].strip()
    path = _os.path.normpath(
        _os.path.join(_os.path.dirname(spec.line.file), local_path)
    )
    if not _os.path.isdir(path):
        logger.debug("%r does not point to a path -> skipping", spec.value)
        return None

    info = None
    for normname, info in installed.items():
        if "editable" not in info:
            continue

        installed_path = _os.path.normpath(info["editable"])
        if installed_path == path:
            logger.debug(
                "Editable installation of %r found at %r", spec.value, path
            )
            return normname

        logger.debug(
            "Skipping editable %r (path does not match)", info["name"]
        )

    logger.debug("Editable %r not found", spec.value)
    return None


def _find_installed(ctx):
    """
    Find installed packages

    Returns:
      dict: norm-name (str) -> {name: str, version: str[, editable: str]}
    """
    found = _json.loads(
        ctx.run(
            ctx.c("pip list --local --format json"),
            echo=False,
            hide=True,
        ).stdout
    )
    name_match = _re.compile(r"^(?:%s)$" % (_parse.regex.idre,)).match

    result = {}
    for pkg in found:
        assert name_match(pkg["name"])
        item = dict(name=pkg["name"], version=pkg["version"])
        if "editable_project_location" in pkg:
            item["editable"] = pkg["editable_project_location"]
        result[_parse.normalize(pkg["name"])] = item

    return result



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Model for req file containing latest versions of dependencies
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import re as _re

from ..._inv import shell as _shell
from .. import _parse

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.suggest.latest_file")


class LatestFile(object):
    """
    Dependency file container

    Attributes:
      name (str):
        The filename
    """

    # pylint: disable = too-many-instance-attributes

    def __init__(self, filename, pattern, prefix, suffix, latest, not_latest):
        """
        Initialization

        Parameters:
          filename (str):
            The filename

          pattern (str):
            Pattern to find the place where to insert the latest deps

          prefix (str):
            Prefix line for the latest deps

          suffix (str):
            Suffix line for the latest deps

          latest (dict):
            normname -> update-info

          not_latest (set):
            Not in latest (possibly anymore)
        """
        # pylint: disable = too-many-positional-arguments

        self.name = _shell.relative(filename)
        self._fullname = _shell.native(filename)
        self._pattern = pattern
        self._prefix, self._suffix = prefix, suffix
        self._add, self._remove = [], []
        self._pre, self._post = None, None
        self._latest, self._not_latest = latest, not_latest
        error = self.load()
        if error:
            logger.debug("%s", error)

    def __repr__(self):
        """
        Create debug representation

        Returns:
          str: The repr of the container
        """
        return "%s(%r, %r, %r, %r, %r, %r)" % (
            self.__class__.__name__,
            self.name,
            self._pattern,
            self._prefix,
            self._suffix,
            self._latest,
            self._not_latest,
        )

    def __str__(self):
        """
        The reconstructed file

        Returns:
          str: The file as string
        """
        pre = "".join(self._pre) + "\n"
        add = "".join(self._add)
        prefix = (
            self._prefix + "\n" if add and self._prefix is not None else ""
        )
        suffix = (
            self._suffix + "\n" if add and self._suffix is not None else ""
        )
        return pre + prefix + add + suffix + self._post

    def as_patch_info(self):
        """
        Create jsonable patch information

        Returns:
          dict: {"add": [], "remove": []} or None
        """
        if set(self._add) == set(self._remove):
            return None
        return dict(add=list(self._add), remove=list(self._remove))

    def as_changed(self):
        """
        Create jsonable patch information

        Returns:
          dict: Name -> new version
        """
        patched = self.as_patch_info()
        if not patched:
            return {}

        removed = {}
        changed = {}
        for line in patched["remove"]:
            req = _parse.requirement(line)
            ops = set(spec.operator for spec in req.specifier)
            versions = [spec.version for spec in req.specifier]
            if ops in ({"=="}, {"~="}):
                removed[_parse.normalize(req.name)] = versions[0]

        for line in patched["add"]:
            req = _parse.requirement(line)
            ops = set(spec.operator for spec in req.specifier)
            versions = [spec.version for spec in req.specifier]
            if ops in ({"=="}, {"~="}):
                name = _parse.normalize(req.name)
                if removed.get(name) != versions[0]:
                    changed[name] = versions[0]

        return changed

    def load(self):
        """
        (Re-)load the file

        Returns:
          str: Error message if any. ``None`` if it was successful.
        """
        try:
            with open(self._fullname, "rb") as fp:
                content = fp.read().decode("latin-1")
        except IOError as e:
            return "Could not read file %r: %s" % (self._fullname, e)

        search = _re.compile("(%s)" % (self._pattern,), _re.M).search
        found = search(content)
        if not found:
            return "Pattern %r not found in %r" % (self._pattern, self.name)

        cut = ""
        pre = content[: found.start(1)]
        post = content[found.start(1) :]
        while pre and not pre.endswith("\n"):
            cut += pre[-1]
            pre = pre[:-1]
        self._pre = pre.splitlines(True)
        self._post = "".join(reversed(cut)) + post

        # Find lines to remove and lines to add
        add = []
        for item in sorted(self._latest.values(), key=lambda x: x["new"]):
            add.append(item["new"].rstrip() + "\n")
        self._add, self._remove = add, self._remove_previous()

        return None

    def _remove_previous(self):
        """
        Remove previous "latest" dependencies from pre

        Returns:
          list: The removed dependencies
        """
        # pylint: disable = too-many-branches

        pre = self._pre
        prefix, suffix = self._prefix or "", self._suffix
        is_req = _re.compile(_parse.regex.full_spec()).match
        remove = []

        while pre:
            # skip empty lines
            if not pre[-1].strip():
                pre.pop()
                continue

            # skip suffix
            if suffix is not None and pre[-1].strip() == suffix.strip():
                pre.pop()
                continue

            # if we hit the prefix, we're done.
            if prefix and pre[-1].strip() == prefix.strip():
                pre.pop()
                break

            # we hit otherwise a line, which does not look like a requirement
            if not is_req(pre[-1]):
                if not prefix:
                    break

                # Scan upwards and look out for the prefix line, skip until
                # then
                idx = 0
                for idx, line in enumerate(reversed(pre), start=1):
                    if line.strip() == prefix.strip():
                        break
                else:
                    break
                for _ in range(idx):
                    pre.pop()
                break

            # Finally we've hit something parsable
            # stop if it's not in our list, except if there's a prefix to be
            # found, then continue scanning
            key = _parse.normalize(_parse.requirement(pre[-1]).name)
            if key not in self._latest and key not in self._not_latest:
                for line in reversed(pre):
                    if line.strip() == prefix.strip():
                        break
                else:
                    break

            remove.append(pre.pop())

        # skip any empty lines
        while pre and not pre[-1].strip():
            pre.pop()
        return remove

    def write(self):
        """Write back the file"""
        # Try to assemble the file before opening it for writing
        result = str(self).encode("latin-1")

        with open(self._fullname, "wb") as fp:
            fp.write(result)



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Inspect package versions in index
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

import json as _json
import logging as _logging

import pkg_resources as _pkg_resources

try:
    from setuptools import package_index as _pkg_index
except ImportError:
    _pkg_index = None
import invoke as _invoke

from ... import pypi as _pypi
from ..._inv import tasks as _tasks
from .. import _parse

# pylint: disable = import-outside-toplevel

logger = _logging.getLogger("deps.inspect.index")


class PipIndex(_pkg_resources.Environment):
    """Query the index using pip"""

    _is_available = None

    def __init__(self, *args, **kwargs):
        super(PipIndex, self).__init__(*args, **kwargs)
        self._ctx = _tasks.new_context()
        self._cache = set()

    @classmethod
    def is_available(cls):
        """Is the pip index available?"""
        if cls._is_available is not None:
            return cls._is_available

        ctx = _tasks.new_context()
        try:
            ctx.run(
                ctx.c(
                    "pip index -i %s versions --json pip",
                    _pypi.index_url(ctx),
                ),
                hide=True,
                echo=False,
            )
        except _invoke.UnexpectedExit:
            cls._is_available = False
        else:
            cls._is_available = True

        return cls._is_available

    def find_packages(self, requirement):
        """Find packages"""
        if requirement.key in self._cache:
            return

        ctx = self._ctx
        index_info = _json.loads(
            ctx.run(
                ctx.c(
                    [
                        "pip",
                        "index",
                        "-i",
                        _pypi.index_url(ctx),
                        "versions",
                        "--json",
                        requirement.project_name,
                    ]
                ),
                hide=True,
            ).stdout.strip()
        )
        for version in index_info.get("versions", ()):
            self.add(
                _pkg_resources.Distribution(
                    project_name=requirement.project_name,
                    version=version,
                    precedence=_pkg_resources.EGG_DIST + 1,
                )
            )
        self._cache.add(requirement.key)

    def obtain(self, requirement, installer=None):
        self.find_packages(requirement)
        for dist in self[requirement.key]:
            if dist in requirement:
                return dist
        return super().obtain(requirement, installer)


class Index(object):
    """Python package index container"""

    def __init__(self):
        """Initialization"""
        if PipIndex.is_available():
            self._index = PipIndex(search_path=[])
        elif _pkg_index is not None:
            self._index = _pkg_index.PackageIndex(search_path=[])
        else:
            raise RuntimeError("No index access tool available")

        self._wset = _pkg_resources.WorkingSet([])

    def lookup(self, req):
        """
        Find the best-fitting requirement in the index

        Parameters:
          req (Requirement):
            The requirement to look up

        Returns:
          Distribution: The found distribution or ``None``
        """
        dist = self._index.best_match(req, self._wset)
        logger.debug("Index lookup result of %r: %r", str(req), dist)
        return dist

    def scanner(self, version):
        """
        Create a new package scanner

        Parameters:
          version (str):
            The package version to compare to

        Returns:
          PackageScanner: New package scanner
        """
        return PackageScanner(self, version)


class PackageScanner(object):
    """Container for scanning a single package"""

    def __init__(self, index, package_version):
        """
        Initialization

        Parameters:
          index (Index):
            The index container

          package_version (str):
            The installed package version
        """
        self._index = index
        self._version = package_version and _parse.version(package_version)
        self._result = {}

    def result(self):
        """
        Get the aggregated result

        Returns:
          dict: The result
        """
        final = dict(self._result)
        if (
            final.get("latest") is not None
            and final.get("compatible") is not None
            and final["latest"] == final["compatible"]
        ):
            del final["latest"]

        if (
            final.get("compatible") is not None
            and final.get("unchanged") is not None
            and final["compatible"] == final["unchanged"]
        ):
            del final["compatible"]

        # if (
        #     final.get("unchanged") is not None
        #     and final["unchanged"] == self._version
        # ):
        #     del final["unchanged"]

        return {
            key: str(value)
            for key, value in final.items()
            if value is not None
        }

    def inspect(self, req):
        """
        Lookup requirement versions in index

        The aggregated result can be retrieved by calling ``.result()``

        Parameters:
          req (Requirement):
            The requirement to look up
        """
        # we want a specifically typed copy
        if PipIndex.is_available():
            req = _pkg_resources.Requirement.parse(str(req))
        else:
            req = _pkg_index.Requirement.parse(str(req))

        self._update_field("unchanged", req)
        self._find_compat(req)
        self._find_latest(req)

    def _find_compat(self, req):
        """
        Find compatible version

        Parameters:
          req (Requirement):
            The requirement to look up
        """
        req = req.parse(str(req))  # copy

        # We turn == into ~=, which finds compatible versions
        # (last digit update)
        specs = []
        for spec in req.specifier:
            operator = spec.operator
            if operator == "==":
                operator = "~="
            specs.append("%s%s" % (operator, spec.version))

        try:
            # pylint: disable = unnecessary-dunder-call
            req.specifier.__init__(",".join(specs), prereleases=False)
        except _pkg_resources.packaging.specifiers.InvalidSpecifier:
            logger.debug("Invalid compat %r", ",".join(specs))
            return

        logger.debug("Valid compat %r", str(req))
        self._update_field("compatible", req)

    def _find_latest(self, req):
        """
        Find latest version

        Parameters:
          req (Requirement):
            The requirement to look up
        """
        req = req.parse(str(req))  # copy

        spec = None if self._version is None else ">=%s" % (self._version,)
        # pylint: disable = unnecessary-dunder-call
        req.specifier.__init__(spec or "", prereleases=False)

        correction = self._update_field("latest", req)
        if correction:
            correction = "< %s" % (correction,)
            spec = ", ".join((spec, correction)) if spec else correction
            req.specifier.__init__(spec or "", prereleases=False)
            self._update_field("latest", req)

    def _update_field(self, field, req):
        """
        Update version info field

        Parameters:
          field (str):
            The field to update

          req (Requirement):
            The requirement to evaluate
        """
        dist = self._index.lookup(req.parse(str(req)))
        if dist is None:
            return None

        dist_version = _parse.version(dist.version)
        if dist_version.is_prerelease:
            return dist_version

        if (self._version is None or dist_version >= self._version) and (
            self._result.get(field) is None
            or dist_version > self._result[field]
        ):
            self._result[field] = dist_version

        return None



# -*- coding: ascii -*-
#
# Copyright 2019 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Containerized Builds
~~~~~~~~~~~~~~~~~~~~

"""

import platform as _platform


class Container(object):
    """Container context"""

    _WAIT_COMMAND = ("/bin/sh", "-c", "while sleep 60; do :; done")

    def __init__(self, ctx, name, image):
        """
        Initialization

        Parameters:
          ctx (Context):
            Context object

          name (str):
            Container name

          image (str):
            Image

          login (callable):
            Login function
        """
        self._ctx = ctx
        # pylint: disable-next = condition-evals-to-constant
        if 0 and _platform.system() == "Darwin" and ctx.which("container"):
            self._command = ContainerCommand(ctx)
        else:
            self._command = DockerCommand(ctx)
        self._image = image
        self._name = name
        self._running = False

    def login(self):
        """Login to registry if needed"""

    def __enter__(self):
        """
        Create packager context

        * Login if requested
        * remove existing container
        * start new container

        Returns:
          Packager: The instance again
        """
        assert not self._running

        self.login()
        self._run(self._command.delete(self._name), warn=True, hide=True)

        self._run(
            self._command.run(
                self._image,
                name=self._name,
                detached=True,
                command=self._WAIT_COMMAND,
            )
        )

        self._running = True
        return self

    def __exit__(self, *args):
        """
        Exit packager context

        * stop the container
        * delete the container
        """
        try:
            kwargs = dict(hide=True, warn=True)
            self._run(self._command.stop(self._name), **kwargs)
            self._run(self._command.delete(self._name), **kwargs)
        finally:
            self._running = False

    def _run(self, command, **kwargs):
        """
        Run command

        Parameters:
          command (list):
            The command to run

        **kwargs:
          Extra arguments for ``ctx.run()``

        Returns:
          ctx.run() response
        """
        ctx = self._ctx
        kwargs.setdefault("echo", True)
        return ctx.run(ctx.c(command), **kwargs)

    def cp_in(self, src, dest):
        """
        cp files into the container

        Parameters:
          src (str):
            Source file or folder

          dest (str):
            Destination in the container
        """
        assert self._running

        self._run(self._command.cp_in(self._name, src, dest))

    def cp_out(self, src, dest):
        """
        cp files out of the container

        Parameters:
          src (str):
            Source file or folder within the container

          dest (str):
            Destination in the local file system
        """
        assert self._running

        self._run(self._command.cp_out(self._name, src, dest))

    def execute(self, command):
        """
        Execute a command in the running container

        Parameters:
          command (list):
            The command to execute
        """
        assert self._running

        self._run(self._command.execute(self._name, command))


class DockerCommand(object):
    """Docker command creator"""

    def __init__(self, ctx):
        """Initialization"""
        self._exe = ctx.which("docker")
        self._ctx = ctx
        self._email_args = (
            ["-e", "none"]
            if "--email"
            in ctx.run(
                ctx.c([self._exe, "login", "--help"]), hide=True
            ).stdout
            else []
        )

    def login(self, url, username):
        """
        Return login command

        Password needs to be passed via stdin

        Parameters:
          url (str):
            Registry URL

          username (str):
            User name

        Returns:
          list: The command
        """
        cmd = [self._exe, "login"] + self._email_args + ["--password-stdin"]
        cmd += ["-u", username]
        cmd += ["--", url]
        return cmd

    def delete(self, container):
        """
        Create command to delete a container

        Parameters:
          container (str):
            The container name

        Returns:
          list: The command
        """
        return [self._exe, "rm", "-f", "--", container]

    def stop(self, container):
        """
        Create command to stop a container

        Parameters:
          container (str):
            The container name

        Returns:
          list: The command
        """
        return [self._exe, "stop", "--", container]

    def run(
        self,
        image,
        name=None,
        detached=None,
        command=None,
        remove=None,
        privileged=False,
    ):
        """
        Create run command

        Parameters:
          image (str):
            The image to run

          name (str):
            Optional container name

          detached (bool):
            Run detached? Default: False

          command (list):
            Optional in-container command to execute

          remove (bool):
            Remove after stop? Default: false

          privileged (bool):
            Run privileged? Default: false

        Returns:
          list: The command
        """
        cmd = [self._exe, "run"]
        if detached:
            cmd += ["-d"]
        if name:
            cmd += ["--name", name]
        if remove:
            cmd += ["--rm"]
        if privileged:
            cmd += ["--privileged"]
        cmd += [image]
        if command:
            cmd += list(command)
        return cmd

    def cp_in(self, container, src, dest):
        """
        Create command to copy files into a container

        Parameters:
          container (str):
            The container name

          src (str):
            Source file or folder

          dest (str):
            Destination in the container

        Returns:
          list: The command
        """
        cmd = [self._exe, "container", "cp"]
        cmd += [src]
        cmd += ["%s:%s" % (container, dest)]
        return cmd

    def cp_out(self, container, src, dest):
        """
        Create command to copy files out of the container

        Parameters:
          src (str):
            Source file or folder within the container

          dest (str):
            Destination in the local file system
        """
        cmd = [self._exe, "container", "cp"]
        cmd += ["%s:%s" % (container, src)]
        cmd += [dest]
        return cmd

    def execute(self, container, command, interactive=True):
        """
        Create command to execute a command in the running container

        Parameters:
          container (str):
            Container name

          command (list):
            The command to execute

          interactive (bool):
            Run interactively? Default: true
        """
        cmd = [self._exe, "exec"]
        if interactive:
            cmd += ["-i"]
        cmd += [container]
        cmd.extend(command)
        return cmd


class ContainerCommand(object):
    """Apple container command creator"""

    def __init__(self, ctx):
        """Initialization"""
        self._exe = ctx.which("container")
        self._ctx = ctx

    def login(self, url, username):
        """
        Return login command

        Password needs to be passed via stdin

        Parameters:
          url (str):
            Registry URL

          username (str):
            User name

        Returns:
          list: The command
        """
        cmd = [self._exe, "registry", "login", "--password-stdin"]
        cmd += ["-u", username]
        cmd += ["--", url]
        return cmd

    def delete(self, container):
        """
        Create command to delete a container

        Parameters:
          container (str):
            The container name

        Returns:
          list: The command
        """
        return [self._exe, "delete", "-f", "--", container]

    def stop(self, container):
        """
        Create command to stop a container

        Parameters:
          container (str):
            The container name

        Returns:
          list: The command
        """
        return [self._exe, "stop", "--", container]

    def run(
        self,
        image,
        name=None,
        detached=None,
        command=None,
        remove=None,
        privileged=None,
    ):
        """
        Create run command

        Parameters:
          image (str):
            The image to run

          name (str):
            Optional container name

          detached (bool):
            Run detached? Default: False

          command (list):
            Optional in-container command to execute

          remove (bool):
            Remove after stop? Default: false

          privileged (bool):
            Run privileged? Default: false


        Returns:
          list: The command
        """
        cmd = [self._exe, "run"]
        if detached:
            cmd += ["-d"]
        if name:
            cmd += ["--name", name]
        if remove:
            cmd += ["--rm"]
        if privileged:
            cmd += ["--privileged"]
        cmd += [image]
        if command:
            cmd += list(command)
        return cmd

    def cp_in(self, container, src, dest):
        """
        Copy files into a container

        Parameters:
          container (str):
            The container name

          src (str):
            Source file or folder

          dest (str):
            Destination in the container

        Returns:
          list: The command
        """
        # TODO: probably wrong
        cmd = [self._exe, "container", "cp"]
        cmd += [src]
        cmd += ["%s:%s" % (container, dest)]
        return cmd

    def cp_out(self, container, src, dest):
        """
        Create command to copy files out of the container

        Parameters:
          src (str):
            Source file or folder within the container

          dest (str):
            Destination in the local file system
        """
        # TODO: probably wrong
        cmd = [self._exe, "container", "cp"]
        cmd += ["%s:%s" % (container, src)]
        cmd += [dest]
        return cmd

    def execute(self, container, command, interactive=True):
        """
        Create command to execute a command in the running container

        Parameters:
          container (str):
            Container name

          command (list):
            The command to execute

          interactive (bool):
            Run interactively? Default: true
        """
        cmd = [self._exe, "exec"]
        if interactive:
            cmd += ["-i"]
        cmd += [container]
        cmd.extend(command)
        return cmd



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Documentation Tasks
~~~~~~~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from .. import _features
from .._inv import tasks as _tasks
from . import _sphinx

NAMESPACE = "doc"


@_tasks.optional(None, _features.sphinx)
@_invoke.task()
def userdoc(ctx):
    """Create userdocs"""
    if _tasks.exists("compile.compile"):
        _tasks.execute(ctx, "compile.compile")

    _sphinx.run(
        ctx, ctx.doc.sphinx.build, ctx.doc.sphinx.source, ctx.doc.userdoc
    )


@_tasks.optional(None, _features.sphinx)
@_invoke.task(userdoc, default=True)
def doc(ctx):  # pylint: disable = unused-argument
    """Create docs"""


@_tasks.optional(None, _features.sphinx)
@_invoke.task()
def website(ctx):
    """Create website"""
    ctx.shell.rm_rf(ctx.doc.website.source)
    ctx.shell.cp_r(
        ctx.doc.sphinx.source, _os.path.join(ctx.doc.website.source, "src")
    )
    ctx.shell.rm_rf(ctx.doc.website.target)
    ctx.shell.mkdir_p(ctx.doc.website.target)

    dlfile = _os.path.join(
        ctx.doc.website.source, "src", "website_download.txt"
    )
    with open(dlfile, "rb") as fp:
        download = fp.read().decode("latin-1")

    ixfile = _os.path.join(ctx.doc.website.source, "src", "index.txt")
    with open(ixfile, "rb") as fp:
        index = fp.read().decode("latin-1").splitlines(True)

    with open(ixfile, "wb") as fp:
        for line in index:
            if line.startswith(".. placeholder: Download"):
                line = download
            fp.write(line.encode("latin-1"))

    _sphinx.run(
        ctx,
        ctx.shell.native(_os.path.join(ctx.doc.website.source, "build")),
        ctx.shell.native(_os.path.join(ctx.doc.website.source, "src")),
        ctx.shell.native(ctx.doc.website.target),
    )
    ctx.shell.rm(_os.path.join(ctx.doc.website.target, ".buildinfo"))



# -*- coding: ascii -*-
#
# Copyright 2018 - 2025
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Requirement manipulation
~~~~~~~~~~~~~~~~~~~~~~~~

"""

import logging as _logging
import re as _re

from .. import _parse

logger = _logging.getLogger("deps.suggest.requirement")


_spec_sub = _re.compile(
    r"""(?x)
        ( (?:[^=]==|[<~]=|<) %(wsp)s* ) ( %(version_str)s )
    """
    % vars(_parse.regex)
).sub


def replace(original, updates, upgrade=False, compat=True, latest=False):
    """
    Find replacement for a single requirement

    Parameters:
      original:
        Original requirement

      updates (dict):
        Available updates for the dependency

      upgrade (bool):
        Try upgrading to the latest version? Default: False

      compat (bool):
        Ensure compatible changes? (``~=``) Default: true

      latest (bool):
        The target is the list of "latest" requirements (e.g. in
        development.txt)

    Returns:
      tuple: new requirement string and new version (extracted)
             (``(str, str)``)
    """
    desired_version = None
    if upgrade:
        desired_version = (
            updates.get("latest")
            or updates.get("compatible")
            or updates.get("unchanged")
        )
    elif not compat:
        desired_version = updates.get("compatible") or updates.get(
            "unchanged"
        )
    if not desired_version:
        return original, None

    version = _parse.version(desired_version)
    if compat:
        replacement = _replace_compat(original, version, upgrade)
    elif latest:
        replacement = _replace_latest(original, version)
    elif upgrade:
        replacement = _replace_upgrade(original, version)
    else:
        replacement = original

    req = _parse.requirement(replacement)
    for spec in req.specifier:
        if spec.operator == "~=" and compat:
            continue
        if spec.operator == "==" and (latest or (compat and not upgrade)):
            continue
        if spec.operator not in ("==", "~=", "<=", "<"):
            continue
        if _parse.version(spec.version) >= version:
            continue

        new_version = plus_1(version) if spec.operator == "<" else version
        replacement = _spec_sub(
            lambda m, n=new_version: "%s%s"
            % (m.group(1), digits(m.group(2), n)),
            replacement,
        )

    return replacement, desired_version


def _replace_compat(requirement, version, upgrade):
    """
    Replace with compat request

    Parameters:
      requirement (str):
        Requirement to manipulate

      version (Version):
        Desired version

      upgrade (bool):
        Is an upgrade requested?

    Returns:
      str: The possibly changed requirement
    """
    req = _parse.requirement(str(requirement))
    ops = set(spec.operator for spec in req.specifier)
    vers = [spec.version for spec in req.specifier]

    if ops == {"~="} and plus_1(digits(vers[0], version)) != plus_1(vers[0]):
        return _spec_sub(
            lambda m: ">= %s, < %s"
            % (
                vers[0],
                digits(vers[0], plus_1(digits(vers[0], version))),
            ),
            requirement,
        )

    if not upgrade:
        return requirement

    if ops == {"=="} and plus_1(version) != plus_1(vers[0]):
        return _spec_sub(
            lambda m: "%s >= %s, <= %s"
            % (
                m.group(1)[0] if m.group(1)[0] not in "=<~> \t" else "",
                vers[0],
                digits(vers[0], version),
            ),
            requirement,
        )

    return _replace_upgrade(requirement, version)


def _replace_upgrade(requirement, version):
    """
    Replace with upgrade request

    Parameters:
      requirement (str):
        Requirement to manipulate

       version (Version):
        Desired version

    Returns:
      str: The possibly changed requirement
    """
    req = _parse.requirement(str(requirement))
    ops = set(spec.operator for spec in req.specifier)

    if ops not in ({">=", "<"}, {">=", "<="}):
        return requirement

    lt_ver = [
        spec.version for spec in req.specifier if spec.operator in ("<", "<=")
    ][0]
    if _parse.version(lt_ver) > version:
        return requirement

    return _spec_sub(
        lambda m: "%s%s"
        % (
            m.group(1),
            digits(
                lt_ver,
                (version if "<=" in ops else plus_1(digits(lt_ver, version))),
            ),
        ),
        requirement,
    )


def _replace_latest(requirement, version):
    """
    Set version to latest

    Parameters:
      requirement (str):
        Requirement to manipulate

       version (Version):
        Desired version

    Returns:
      str: The possibly changed requirement
    """
    req = _parse.requirement(str(requirement))
    ops = set(spec.operator for spec in req.specifier)

    if ops not in ({">=", "<"}, {">=", "<="}):
        return requirement

    return _re.sub(
        _parse.regex.full_spec(req.name),
        lambda m: "%s%s ~= %s%s%s"
        % (
            m.group("name"),
            m.group("extra"),
            digits(m.group("version"), version),
            m.group("version")[len(m.group("version").rstrip()) :],
            m.group("marker"),
        ),
        requirement,
    )


def digits(ref_version, target_version):
    """
    Ensure significant number of version number items

    Parameters:
      ref_version (str):
        Reference version

      target_version (str):
        The version to adjust

    Returns:
      str: target version with possibly adjusted number of digits
    """
    release_match = _re.compile(_parse.regex.version_release).search

    ref_release = release_match(str(ref_version))
    if not ref_release:
        return target_version

    target_release = release_match(str(target_version))
    if not target_release:
        return target_version

    no_digits = len(ref_release.group(2).split("."))
    return "%s%s" % (
        target_release.group(1),
        ".".join(
            (target_release.group(2).split(".") + (["0"] * no_digits))[
                :no_digits
            ]
        ),
    )


def plus_1(version, which=-2):
    """
    Increase version number, so the < operator can be applied

    Parameters:
      version (Version or str):
        The version to increase

      which (int):
        Which version index to change. If it's outside the range of the
        versions, -1 is applied. Default: -2

    Returns:
      str: Increased version as a string
    """
    # group(1) is optional "epoch", group(2) the actual numbered version items
    # group(1) will be passed through, group(2) will be manipulated
    release = _re.match(_parse.regex.version_release, str(version))
    assert release, repr(str(version))

    numbers = [int(item) for item in release.group(2).split(".")]
    index = len(numbers) + which if which < 0 else which
    if not 0 <= index < len(numbers):
        index = len(numbers) - 1

    numbers[index] += 1
    return "%s%s" % (
        release.group(1),
        ".".join(map(str, numbers[: index + 1])),
    )
